//
// 
//

#include "Model.h"
#include "GlobalData.h"

Model::Model(SequDatabase& db) : db(db) {
    this->db = db;
    //init cover, start with all singletons in model
    sum_of_usages = 0;
    for (int e = 0; e <= db.getHighestEvent();e++){
        Pattern* p = new Pattern(e,{e}, {-1},{NONE});
        selection[e] = p;
    }
    patternCounter = db.getHighestEvent(); //id of last added event
    abstractionCounter = db.getHighestEvent(); //id of last created abstraction - starting at highest event id


    for (auto  id_pattern: selection){
        auto p_windows = findWindows(*(id_pattern.second), GlobalData::window_budget);
        p_windows.sort();
        availableWindowSet.addToWindowSet(p_windows);
    }
    cover = greedyCover();
    recomputeUsages();
    computeFixedCost();
    computeLength();

    usage_copy_diagnositics = pattern_singleton_usage;
}

Model::~Model() {
    for (auto p: selection){
        delete p.second;
    }
}




void Model::computeLength(bool ignore_spec_cost) {
    double ct_length = computeCodeTableLength();
    double cover_length = computeCoverLength(ignore_spec_cost);
    encoded_length =  ct_length + cover_length + fixed_cost;

    this->ct_length = ct_length;
    this->cover_length = cover_length;
    this->fix_length = fixed_cost;
//    std::cout << "CT Length: " << ct_length << std::endl;
//    std::cout << "Cover Length: " << cover_length << std::endl;
}

double Model::computeCoverLength(bool ignore_spec_cost) {
    const double epsilon = 0.5;
    double bits  = 0;
    double d_sum = (double)sum_of_usages;
    for (auto patternId_usage:pattern_singleton_usage){ // compute bits needed for data sequence
        if (patternId_usage.second == 0){
            throw std::runtime_error("Used pattern not allowed!"
                                     "Likely input file error\nMake sure to start with event 0");
        }
        double d_usage = (double)patternId_usage.second;
        bits += d_usage * -log2(d_usage/d_sum);
    }
    double data_sequ_bits = bits;
//    std::cout << "Data Sequ Bits: " << data_sequ_bits <<  std::endl;
    for (auto patternId_gaps:gap_counts){
        int fills = pattern_singleton_usage[patternId_gaps.first] * (selection[patternId_gaps.first]->getEvents().size()-1); //-1 because first fill code is implicit
        for (int i =0; i < fills;i++){
            bits += log2((double)(2*epsilon + i) / ((double)(i+epsilon)));
        }
        for (int i = 0; i < patternId_gaps.second; i++){
            bits += log2(((double)(2*epsilon + i + fills))/((double)(i+epsilon)));
        }
    }
    double meta_sequ_bits = bits - data_sequ_bits;
//    std::cout << "Meta Sequ Bits " << meta_sequ_bits << std::endl;
    if (!ignore_spec_cost){
        for (double p:specificationProbabilities){
            bits += -log2(p);
        }
    }
    double spec_sequ_bits = bits - meta_sequ_bits - data_sequ_bits;
//    std::cout << "Spec Sequ Bits: " << spec_sequ_bits << std::endl;
    return bits;
}

double Model::computeCodeTableLength() {
    double bits = 0;
    bits += universal_integer_encoding(abstractions.size() +1); //encode number of abstractions
    std::vector<int> k_abstraction_in_abstraction;
    int abs_counter = 0;
    for (auto absId_abstraction : abstractions){ //encode abstractions
        k_abstraction_in_abstraction.push_back(otherAbstractions(absId_abstraction.second));
        if (abs_counter != 0){
            bits += log2(abs_counter);
        }
        abs_counter++;
        int m = absId_abstraction.second.getChildren().size() - otherAbstractions(absId_abstraction.second);
        int m_possible_children = db.getHighestEvent()-getAllChildrenOfAbstraction(absId_abstraction.second).size() + m; // get all children incl. m hence has to be added back in
        bits += log2(m_possible_children); //encodes m
        bits += logBinom(std::max(m_possible_children,0), std::max(m,0));
    }
    std::sort(k_abstraction_in_abstraction.begin(), k_abstraction_in_abstraction.end());
    for (int i = 1;i<k_abstraction_in_abstraction.size();i++){
        bits += logBinom(i, k_abstraction_in_abstraction[i]);
    }

    //encode number and usage of patterns
    bits += universal_integer_encoding(selection.size() - db.getHighestEvent()+1); //+1 in case there are no patterns
    bits += universal_integer_encoding(sum_of_usages);
    bits += logBinom(sum_of_usages+db.getHighestEvent()+1 -1, selection.size()-1);
    for (auto id_pattern : selection){
        if ((id_pattern.second)->getEvents().size() > 1){ // only encode "real" patterns
            bits += computePatternLength(*(id_pattern.second));
        }
    }
    return bits;
}

void Model::computeFixedCost() { //computes cost that are fixed for one dataset.
    double bits = 0;
    // number of unique events aka. alphabet size
    bits += universal_integer_encoding(db.getHighestEvent());
    // fixed cost in cover
    bits += universal_integer_encoding(db.size()); // number of sequences
    for (int i = 0; i < db.size(); i++){ // length of each sequence
        bits += universal_integer_encoding(db.getSequence(i).size());
    }
    fixed_cost = bits;
}

double Model::getEncodedLength() const {
    return encoded_length;
}

double Model::computePatternLength(const Pattern &pattern) {
    double bits = 0;
    int pattern_length = pattern.getEvents().size();
    bits += universal_integer_encoding(pattern_length);
    bits += (double)pattern_length * -log2(1.0/(double )(db.getHighestEvent()+1 + abstractions.size())); //uniform cost over all events and abstractions
    int unique_abstractions = 1; // fist option is -1
    for (int i = 0; i < pattern.getEvents().size(); i++){
        auto symbol = pattern.getEvents()[i]; //symbol can either be event or abstraction
        if (symbol > db.getHighestEvent()){ // if symbol is an abstraction
            bits += log2(unique_abstractions); //encodes pattern symbol it dependents on
            if (pattern.kindOfDependence(i) != NONE){ // sanity check to catch invalid patterns - could be removed
                if (pattern.dependentOn(i) == -1){ // depends on no other symbol of pattern implies DEPEND_ON link
                    unique_abstractions++;
                    // no further action required here - symbol is fully specified
                }else{
                    bits += log2(static_cast<int>(LinkType::count)-1); // encodes what kind of relation - none does not need to be considered
                    if (pattern.kindOfDependence(i) != EQUALS){
                        unique_abstractions++;
                    }
                }
            }else{
                throw std::runtime_error{"Invalid Pattern Structure, can't have abstraction with NONE link type!"};
            }
        }
    }
    return bits;
}


WindowSet Model::findWindows(Pattern &pattern, int budget) {
    return findWindows(pattern,budget,AbstractionExtension(),{});
}

WindowSet Model::findWindows(Pattern &pattern, int budget, AbstractionExtension current_abs_ext,
                             std::vector<AbstractionExtension> prev_abs_extensions) {
    pattern.clearWindows_and_Contingency_tables();
    WindowSet windows;
    std::priority_queue<PQEntryFindWin> priorityQueue; 

    auto pattern_events = pattern.getEvents();
    std::vector<std::vector<int>> all_children(pattern.getEvents().size());
    for (int i =0; i < pattern_events.size();i++){
        all_children[i] = getAllChildrenOfEvent(pattern_events[i], current_abs_ext, prev_abs_extensions, {});
    }

    auto possible_first_symbols = getAllChildrenOfEvent(pattern.getEvents()[0], current_abs_ext, prev_abs_extensions,
                                                        {});
    for (int event:possible_first_symbols){
        for (DataIndex location: db.getLocationsOfEvent(event)){
            if (1 == pattern.getEvents().size()){ // if singleton pattern add directly to windowset - while loop will be skipped since nothing gets added to pq
                Window w = Window(&pattern, {location});
                windows.push_back(w);
                pattern.addWindow(w);
            } else{
                priorityQueue.push(PQEntryFindWin(location,pattern.getEvents().size()));
            }
        }
    }
    while (!priorityQueue.empty()){
        PQEntryFindWin top_entry = priorityQueue.top();
        priorityQueue.pop();
        top_entry.current_index.event_id++;
        sequence data_sequence = db.getSequence(top_entry.current_index.sequence_id);
        if (top_entry.current_index.event_id >= data_sequence.size()){
            // pattern not found before sequence end
            continue;
        }
        bool added_position = false;
        for (int event: all_children[top_entry.getNextEventIndex()]){
            if (event == data_sequence[top_entry.current_index.event_id]){
                auto top_entry_copy = top_entry;
                top_entry_copy.gaps++;
                if (top_entry_copy.gaps <= pattern.getEvents().size()*budget){
                    priorityQueue.push(top_entry_copy);
                }
                top_entry.addPosition(top_entry.current_index);
                added_position = true;
                if (top_entry.getNextEventIndex() == pattern.getEvents().size()){
                    // pattern completely found in data
                    Window w = Window(&pattern, top_entry.getPositions());
                    windows.push_back(w);
                    pattern.addWindow(w);
                } else{
                    priorityQueue.push(top_entry); // add updated sill incomplete window back into pq.
                }
                break;
            }
        }
        if (added_position){
            continue; // successfully extended pattern by one symbol
        }
        // inc gaps if no symbol was found
        top_entry.gaps++;
        if (top_entry.gaps <= pattern.getEvents().size()* budget){ // + top_entry.pattern_length // rethink budget approach - when exactly do we want to discard a window?
            priorityQueue.push(top_entry);
        }
        
    }
    pattern.allWindowsSet();
    windows.sort();
    return windows;
}

std::vector<int> Model::getAllChildrenOfAbstraction(Abstraction& abstraction) {
    std::set<int> all_children;
    for (int child:abstraction.getChildren()){
        if (child > db.getHighestEvent()){ // abstraction ids start with highest event id +1
            if (child == abstraction.getId()){
                throw std::runtime_error("Abstraction contains itself");
            }
            auto children_of_nested_abstraction = getAllChildrenOfAbstraction(getAbstraction(child));
            all_children.insert(children_of_nested_abstraction.begin(), children_of_nested_abstraction.end());
        } else{
            all_children.insert(child);
        }
    }
    std::vector<int> all_child_vec;
    all_child_vec.insert(all_child_vec.end(),all_children.begin(), all_children.end());
    return all_child_vec;
}

std::vector<int> Model::getAllAbstractionsOfEvent(int event){
    std::vector<int> all_abstraction;
    if (event > db.getHighestEvent()){
        for (auto child:getAbstraction(event).getChildren()){
            auto all_abs_of_child = getAllAbstractionsOfEvent(child);
            all_abstraction.insert(all_abstraction.end(),all_abs_of_child.begin(),all_abs_of_child.end());
        }
        all_abstraction.push_back(event);
    }
    return all_abstraction;
}

std::vector<int> Model::getAllAbstractionsOfEvent(int event, AbstractionExtension ext, std::vector<AbstractionExtension> extensions){
    std::vector<int> all_abstraction;
    extensions.push_back(ext);
    if (event > db.getHighestEvent()){
        bool in_ext =false;
        for (auto ext:extensions){
            if (ext.id == event){
                in_ext = true;
                for (int child:ext.events){
                    auto abs_of_abs = getAllAbstractionsOfEvent(child,AbstractionExtension(),extensions);
                    all_abstraction.insert(all_abstraction.end(),abs_of_abs.begin(), abs_of_abs.end());
                }
            }
        }
        if (in_ext){
            all_abstraction.push_back(event);
        }else{
            return getAllAbstractionsOfEvent(event);
        }

    }
    return all_abstraction;
}

std::vector<int> Model::getAllChildrenOfEvent(int event) {
    if (event > db.getHighestEvent()){
        return getAllChildrenOfAbstraction(getAbstraction(event));
    }else{
        return {event};
    }
}

std::vector<int>
Model::getAllChildrenOfEvent(int event, AbstractionExtension latest_ext, std::vector<AbstractionExtension> extensions,
                             std::set<int> seen, bool ignore_self) {
    std::vector<int> all_children;
    extensions.push_back(latest_ext); // extension is local copy , no sideeffects
    bool in_ext = false;
    seen.insert(event);
    for (auto ext:extensions){ //incl. extensions
        if (ext.id == event){
            for (auto child:ext.events){
                if (seen.find(child) == seen.end()){
                    auto all_children_of_child = getAllChildrenOfEvent(child, {}, extensions, seen); //why does this work??
                    all_children.insert(all_children.end(),all_children_of_child.begin(),all_children_of_child.end());
                }

            }
            if (ext.action == ADD_NEW){
                in_ext = true;
            }
        }
    }
    if (!in_ext){
        if (event > db.getHighestEvent()){
            for (auto child: getAbstraction(event).getChildren()){ //getChildren gives only direct children
                if (seen.find(child) == seen.end()) {
                    auto all_children_of_child = getAllChildrenOfEvent(child, {}, extensions, seen);
                    all_children.insert(all_children.end(), all_children_of_child.begin(), all_children_of_child.end());
                }
            }
        }else if(!ignore_self){
            all_children.push_back(event);
        }
    }
    return all_children;
}

std::vector<int>
Model::getAllChildrenOfAbsExt(AbstractionExtension ext, std::vector<AbstractionExtension> abst_ext_list) {
    std::vector<int> all_children;
    for (int e:ext.events){
        auto all_children_of_e = getAllChildrenOfEvent(e, {}, abst_ext_list, {});
        all_children.insert(all_children.end(),all_children_of_e.begin(),all_children_of_e.end());
    }
    return all_children;
}

Abstraction& Model::getAbstraction(int i) {
    auto findReturn = abstractions.find(i);
    if (findReturn == abstractions.end()) {
        throw std::runtime_error("Abstraction with id not found!");
    }else{
        return findReturn->second;
    }
}
WindowSet Model::greedyCover() {
    WindowSet empty;
    return greedyCover(empty);
}
WindowSet Model::greedyCover(WindowSet& candidate_windows) {




    WindowSet mergedWindows = availableWindowSet.getAugmentedWindowSet(candidate_windows);

    for (auto it = mergedWindows.begin(); it != mergedWindows.end();){
        auto conflictSearcher = it;
        auto greatestWindow = it;
        if (conflictSearcher != mergedWindows.end()) conflictSearcher++;
        while (conflictSearcher != mergedWindows.end() && greatestWindow->getLastPosition() >= conflictSearcher->getFirstPosition()){
            if(checkOverlap(greatestWindow->getPositions(), conflictSearcher->getPositions()) && l_greater_r(&*conflictSearcher, &*greatestWindow)){
                greatestWindow = conflictSearcher;
            }
            conflictSearcher++;
        }
        if (greatestWindow == it) {
            it++;
        }
        else if (checkOverlap(greatestWindow->getPositions(), it->getPositions()) && l_greater_r(&*greatestWindow, &*it)){
            it = mergedWindows.erase(it);
        }
        auto kickOutIterator = it;
        bool kickouteqit = true;
        while (kickOutIterator != mergedWindows.end() && greatestWindow->getLastPosition() >= kickOutIterator->getFirstPosition()){
            if (greatestWindow != kickOutIterator && checkOverlap(greatestWindow->getPositions(), kickOutIterator->getPositions()) && l_greater_r(&*greatestWindow, &*kickOutIterator)){
                kickOutIterator = mergedWindows.erase(kickOutIterator);
                if(kickouteqit) it =kickOutIterator;
            }else{
                kickOutIterator++;
                kickouteqit = false;
            }
        }
    }


    return mergedWindows;
}

bool l_greater_r(Window* lhs, Window* rhs){
    auto lhs_events = lhs->getPattern()->getEvents();
    auto rhs_events = rhs->getPattern()->getEvents();
    if (lhs_events.size() == rhs_events.size()){
        auto lhs_w_length = lhs->getLastPosition() - lhs->getFirstPosition();
        auto rhs_w_length = rhs->getLastPosition() - rhs->getFirstPosition();
        if (lhs_w_length == rhs_w_length){
            auto lhs_support = lhs->getPattern()->allWindowsSize();
            auto rhs_support = rhs->getPattern()->allWindowsSize();
            if (lhs_support == rhs_support){
                int i =0;
                while (lhs_events[i] == rhs_events[i] && i < lhs_events.size()){
                    i++;
                }
                return lhs_events[i] < rhs_events[i];
            }else{
                return lhs_support > rhs_support;
            }
        }else{
            return lhs_w_length < rhs_w_length;
        }
    }else {
        return (lhs_events.size() > rhs_events.size());
    }

}

std::vector<Window*> Model::conflictResolve(std::vector<Window *> conflictSet) {
    std::sort(conflictSet.begin(), conflictSet.end(), l_greater_r);
    return conflictSet;
}

void Model::recomputeUsages() {
    pattern_singleton_usage.clear();
    gap_counts.clear();
    specificationProbabilities.clear();


    for (auto id_pattern:selection){
        id_pattern.second->clearUsageCounts();
        id_pattern.second->clearLocations();
    }

    for (Window& w:cover){
        int patternId = w.getPattern()->getId();
        auto it = pattern_singleton_usage.find(patternId);
        if (it == pattern_singleton_usage.end()){
            pattern_singleton_usage[patternId] =1;
            gap_counts[patternId] = w.getGapCount();
        }else{
            it->second++;
            gap_counts[patternId] += w.getGapCount();
        }

        w.getPattern()->addLocation(&w);

        for (int i = 0; i < w.getPositions().size(); i++){
            DataIndex di = w.getPositions()[i];
            int covered_event = db.getEvent(di);
            int pattern_event = w.getPattern()->getEvents()[i];
            if (pattern_event != covered_event){
                if (w.getPattern()->kindOfDependence(i) != EQUALS){
                    // pattern event has to be non instantiated abstraction -> specification required
                    if (w.getPattern()->kindOfDependence(i) == DEPENDENT_ON){
                        int conditionedBy;
                        if (w.getPattern()->dependentOn(i) == -1){
                            conditionedBy = -1;
                        }else{
                            conditionedBy = db.getEvent(w.getPositions()[w.getPattern()->dependentOn(i)]);
                        }
                        int usg_count = w.getPattern()->getUsageCount(i, conditionedBy, covered_event);
                        int sum_usage_count = w.getPattern()->getSumUsageCount(i,conditionedBy);
                        int n_options = getAllChildrenOfEvent(pattern_event).size();
                        double epsilon = 0.5;
                        double specification_prob = (epsilon + usg_count) / (n_options*epsilon + sum_usage_count);
                        specificationProbabilities.push_back(specification_prob);
                        w.getPattern()->incUsageCounts(i,conditionedBy,covered_event);
                    }
                    //else case schould not happen
                }

            }
        }
    }
    sum_of_usages = cover.size(); 

}



void Model::addPatternToSelection(Pattern *p, bool assign_id) {
    if (assign_id){
        patternCounter++;
        p->setId(patternCounter);
        selection[patternCounter] = p;
    }else{
        selection[p->getId()] = p;
    }

}

int Model::otherAbstractions(Abstraction& abstraction) {
    int count = 0;
    std::set<int> seen;
    for (int c:abstraction.getChildren()){
        if (c > db.getHighestEvent()){
            if (seen.find(c) == seen.end()){ 
                count++;
                seen.insert(c);
            }

        }
    }
    return count;
}

int Model::addAbstraction(Abstraction a, bool assignId) {
    int id = (assignId) ? ++abstractionCounter : a.getId();
    if (assignId) a.setId(id);
    abstractions.insert({id,a});
    return id;
}

void Model::extendAbstraction(int id, std::vector<int> new_events) {
    auto abs = abstractions.find(id);
    for (auto e:new_events){
        (*abs).second.extendAbstraction(e);
    }
}

void Model::removeElementFromAbstraction(int abs_id, int element) {
    auto abs = abstractions.find(abs_id);
    abs->second.removeChildrenFromAbstraction({element});
}

int Model::getNewAbstractionId() {
    return ++abstractionCounter;
}

void Model::removeAbstraction(int id) {
    abstractions.erase(id);
}

WindowSet &Model::getCover() {
    return cover;
}

void Model::loadAbstractionsFromFile(const std::string& filePath) {
    FILE* f = fopen(filePath.c_str(), "r");
    std::vector<int> firstChildSet;
    std::vector<int>& children = firstChildSet;
    int ed;
    while (fscanf(f,"%d", &ed) == 1){
        if (ed == -1){
            addAbstraction(Abstraction(children), true);
            children = std::vector<int>();
        } else{
            children.push_back(ed);
        }
    }

}

void Model::loadPatternsFromFile(const std::string& filePath){
    FILE* f = fopen(filePath.c_str(), "r");
    std::vector<int> firstEvents;
    std::vector<int>& events = firstEvents;
    std::vector<int> firstDependOn;
    std::vector<int>& dependOn = firstDependOn;
    std::vector<LinkType> firstLinkType;
    std::vector<LinkType>& linkType = firstLinkType;

    enum ReadMode {EVENTS, DEPEND, LINK};

    ReadMode state = EVENTS;

    int ed;
    while (fscanf(f,"%d", &ed) == 1){
        if (ed == -1){ //pattern is over
            state = EVENTS;
            Pattern* nextPattern = new Pattern(events,dependOn,linkType);
            addPatternToSelection(nextPattern, true);
            auto p_windows = findWindows(*nextPattern,GlobalData::window_budget);
            p_windows.sort();
            availableWindowSet.addToWindowSet(p_windows);
            events = std::vector<int>();
            dependOn = std::vector<int>();
            linkType = std::vector<LinkType>();
        }
        else if (ed == -2){
            state = static_cast<ReadMode>(state+1);
        }
        else if (state == EVENTS){
            events.push_back(ed);
        }
        else if (state == DEPEND){
            if (ed == -3){
                dependOn.push_back(-1);
            }else{

                dependOn.push_back(ed);
            }
        } else if (state == LINK){
            linkType.push_back(static_cast<LinkType>(ed));
        }
    }


    // recompute cover and length with newly added patterns
    cover = greedyCover();
    recomputeUsages();
    computeLength();
}

void Model::loadOtherPatternsFromFile(const std::string& filePath) {
    FILE* f = fopen(filePath.c_str(), "r");

    std::vector<int> firstEvents;
    std::vector<int>& events = firstEvents;
    std::vector<int> dependent_on;
    std::vector<LinkType> link_type;


    int ed;
    while (fscanf(f,"%d", &ed) == 1){
        if (ed == -1){
            Pattern* nextPattern = new Pattern(events,dependent_on,link_type);
            addPatternToSelection(nextPattern, true);
            auto p_windows = findWindows(*nextPattern,GlobalData::window_budget);
            p_windows.sort();
            availableWindowSet.addToWindowSet(p_windows);

            events = std::vector<int>();
            dependent_on.clear();
            link_type.clear();

        }else{
            events.push_back(ed);
            dependent_on.push_back(-1);
            link_type.push_back(NONE);
        }
    }
    cover = greedyCover();
    recomputeUsages();
    computeLength();
    std::pair<std::vector<int>, std::vector<std::pair<int,int>>> merge_out;
    do {
        merge_out = mergePatternsAndFlattenAbs();
    }while(!merge_out.first.empty() || !merge_out.second.empty());
    removeUnusedAbstractions();

}

void Model::computePatternStructureForAll(){
    for (auto p:selection){
        if (p.first > GlobalData::db->getHighestEvent()){
            p.second->copyLocations();
            computePattenStructure(p.second);
        }
    }
    recomputeUsages();
    computeLength();
}

inline void addToDistribution(std::map<int,GapDistribution>& distributions, WindowSet::iterator& coverIt, int difference){
    int patternId = coverIt->getPattern()->getId();
    if (distributions.find(patternId) != distributions.end()){
        distributions[patternId].incDeltaFrequency(difference,1);
    }else{
        distributions[patternId] = GapDistribution();
        distributions[patternId].incDeltaFrequency(difference,1);
    }
}

Candidate Model::extendByFollowing(Pattern &pattern) { 
    auto coverIt = cover.begin();
    std::map<int,GapDistribution> distributions;
    for (auto patternLocation:pattern.getLocations()){ // locations in pattern have same order as in cover
        while (&*coverIt != patternLocation){
            coverIt++;
        }
        //coverIt and patternLoc point to same window
        std::set<int> seen;
        coverIt++;
        // && coverIt->getFirstPosition() - patternLocation->getLastPosition() < GlobalData::window_budget
        while (coverIt != cover.end() && patternLocation->getLastPosition().sequence_id == coverIt->getLastPosition().sequence_id && *(coverIt->getPattern()) != pattern && (coverIt->getFirstPosition() - patternLocation->getLastPosition() < pattern.getEvents().size() * GlobalData::window_budget || GlobalData::disable_window_in_extension_search)){ // while in same sequence
            // how is the cover ordered?
            if (seen.find(coverIt->getPattern()->getId()) == seen.end()){
                if (coverIt->getFirstPosition() > patternLocation->getLastPosition()){
                    seen.insert(coverIt->getPattern()->getId());
                    addToDistribution(distributions,coverIt,coverIt->getFirstPosition() - patternLocation->getLastPosition());
                }
            }
            coverIt++;

        }
        if (coverIt != cover.end() && *(coverIt->getPattern()) == pattern && patternLocation->getLastPosition().sequence_id == coverIt->getLastPosition().sequence_id){
            if (coverIt->getFirstPosition() > patternLocation->getLastPosition()){
//                addToDistribution(distributions,coverIt,coverIt->getFirstPosition() - patternLocation->getLastPosition());
            }
        }
    }
    return Candidate(pattern,pattern.getEvents().size(),distributions);
}

Candidate Model::extendByProceeding(Pattern &pattern) {
    auto coverIt = cover.begin();
    std::map<int,GapDistribution> distributions;
    for (auto pattenLocation:pattern.getLocations()){
        while (&*coverIt != pattenLocation){
            coverIt++;
        }
        if (coverIt != cover.begin()){
            coverIt--;
        }
        while(pattenLocation->getFirstPosition() < coverIt->getLastPosition() && coverIt != cover.begin()){ // since cover is ordered on last position
            coverIt--;
        }
        std::set<int> seen;
        // && pattenLocation->getFirstPosition() - coverIt->getLastPosition() < GlobalData::window_budget
        while (coverIt != cover.begin() &&  pattenLocation->getLastPosition().sequence_id == coverIt->getLastPosition().sequence_id && *(coverIt->getPattern()) != pattern && (pattenLocation->getFirstPosition() - coverIt->getLastPosition() <   pattern.getEvents().size() * GlobalData::window_budget || GlobalData::disable_window_in_extension_search)){
            if (seen.find(coverIt->getPattern()->getId()) == seen.end()){
                seen.insert(coverIt->getPattern()->getId());
                addToDistribution(distributions,coverIt, coverIt->getLastPosition() - pattenLocation->getFirstPosition());
            }
            coverIt--;
        }
        if ((coverIt == cover.begin()) ||
             pattenLocation->getLastPosition().sequence_id == coverIt->getLastPosition().sequence_id && *(coverIt->getPattern()) == pattern){
//            addToDistribution(distributions,coverIt,coverIt->getLastPosition() - pattenLocation->getFirstPosition());
        }
    }
    return Candidate(pattern,0,distributions);
}

std::vector<std::map<int,GapDistribution>> Model::extendInterleaving(Pattern &pattern) { // element that frequently occur between
    auto p_size = pattern.getEvents().size();
    std::vector<std::map<int,GapDistribution>> counts_after_i = std::vector<std::map<int,GapDistribution>>(p_size-1);
    for (auto location:pattern.getLocations()){
        auto last_pos = location->getFirstPosition();
        int i = -1;
        for (auto pos:location->getPositions()){
//            if (i >= p_size-1) break;
            auto current_position = last_pos;
            ++current_position;
            while (current_position < pos){
                auto event = db.getEvent(current_position);
                auto e_it = counts_after_i[i].find(event);
                if (e_it != counts_after_i[i].end()){
                    e_it->second.incDeltaFrequency(current_position - last_pos,1);
                }else{
                    counts_after_i[i].insert({event,GapDistribution()});
                    counts_after_i[i][event].incDeltaFrequency(current_position - last_pos,1);
                }
                ++current_position;
            }
            i++;
            last_pos = pos;
        }
    }
    return counts_after_i;
}

patternSet Model::getSelection() {
    return selection;
}




AbstractionExtension Model::extendPattern(PatternExtension &p, int position, int extension, bool extend_abstraction,std::vector<AbstractionExtension> abstraction_extension_list) {
    int new_event = extension;
    AbstractionExtension abstractionExtension;
    if (extend_abstraction){
        LocationExtension current_event = p.getLocationExtension(position); //p.getEvents()[position];
        if (current_event.event <= db.getHighestEvent()){ // non abs element -> new abs needs to be created
            //add current and extension into new abstraction
            new_event = getNewAbstractionId();
            abstractionExtension = AbstractionExtension(ADD_NEW,{current_event.event, extension}, new_event);


        }else{ // already an abstraction -> extend current abstraction
            new_event = current_event.event;

            //check if current event in abs_ext_list
            std::vector<int> all_children;
            bool children_found = false;
            for (auto ext:abstraction_extension_list){
                if (ext.id == current_event.event){
                    auto all_children_ext = getAllChildrenOfAbsExt(ext, abstraction_extension_list);
                    all_children.insert(all_children.end(),all_children_ext.begin(),all_children_ext.end());
                    if (ext.action == ADD_NEW){
                        children_found = true;
                    }
                }
            }
            if (!children_found){ // must be abs already added to model
                auto all_children_event = getAllChildrenOfEvent(current_event.event);
                all_children.insert(all_children.end(),all_children_event.begin(),all_children_event.end());
            }
            bool already_in_abs = false;
            for (int e:all_children){
                if (e == extension){
                    already_in_abs = true;
                }
            }
            if (!already_in_abs){

                auto frequUpdate = UniqueValueCounter<int>();
                // key: <pattern-id, abs_loc_in_pattern> value: vector<int> added elements with freuqu
                std::map<std::pair<int,int>, std::vector<int>> absUpdates;
                // find affected patterns
                for (auto p :selection){
                    int e_i = 0;
                    for (auto e:p.second->getEvents()){

                        bool affected = false;
                        for (auto child:getAllAbstractionsOfEvent(e)){
                            if(child == current_event.event){
                                affected = true;
                                break;
                            }
                        }
                        if (affected){
                            auto tmpPattern = Pattern(*p.second);
                            // find windows applies to all e_i

                            auto windowList = findWindows(tmpPattern, GlobalData::window_budget,abstractionExtension, abstraction_extension_list);// find windows with abstaction extension applied
                            int patten_increase = windowList.size()-selection[p.first]->allWindowsSize();
                            frequUpdate.incCount(p.first,patten_increase);
                            // other patterns usage decreases -> assume to be covered by singletons
                            for (int i = 0;i <tmpPattern.getEvents().size(); i++){
                                auto e = tmpPattern.getEvents()[i];
                                if (e <= db.getHighestEvent()){
                                    frequUpdate.incCount(e,-patten_increase);
                                }
                                for (auto c:getAllChildrenOfEvent(e)){
                                    double c_counts = p.second->getUnconditionalUsageCounts(i).getCount(c);
                                    double p_usage = p.second->getPatternUsageCounts();
                                    auto abs_diff = (p_usage == 0)? 0 :c_counts/p_usage * patten_increase;
                                    frequUpdate.incCount(c,-abs_diff);
                                }
                            }

                            for (auto extendedElement :getAllChildrenOfEvent(extension)){
                                int diff = tmpPattern.getMarginCounts(e_i,extendedElement);
                                frequUpdate.incCount(extendedElement,-diff);
                            }
                            break;
                        }
                        e_i++;
                    }
                }
                abstractionExtension = AbstractionExtension(EXTEND, {extension},current_event.event,frequUpdate,absUpdates);
            }// else it leaves abstactionExtension as no nothing
        }
        p.addExtension(new_event,position,KIND_REPLACE);
    } else{ // splice in ?
        p.addExtension(new_event,position,KIND_SPLICE);
    }
    return abstractionExtension;
}

ExtendCandidateReturn
Model::extendPattern(Pattern &pattern, std::vector<std::map<int, GapDistribution>>& counts_after_i) {
    std::vector<pq_gap_dis_ext> extension_queues = std::vector<pq_gap_dis_ext>(counts_after_i.size());
    std::vector<std::pair<int,int>> i_order;
    for (int i = 0; i < counts_after_i.size(); i++){
        int max_total_count =0;
        for (auto e_gap_ds:counts_after_i[i]){
            auto total_counts = e_gap_ds.second.getTotalCounts();
            if (total_counts > 1){

                extension_queues[i].emplace(i,e_gap_ds.first,-e_gap_ds.second.getTotalCounts());
            }
            if (total_counts > max_total_count){
                max_total_count = total_counts;
            }
        }
        i_order.emplace_back(i,max_total_count);
    }

    std::sort(i_order.begin(), i_order.end(), [](auto &left, auto &right) {
        return left.second > right.second;
    });


    std::vector<AbstractionExtension> abs_ext_list;
    int abs_ext_best_i = 0;
    AbstractionExtension abs_ext;
    auto pExt = PatternExtension(pattern, true);
    auto bestPExt = pExt;

    double highest_gain = -DBL_MAX;
    double last_gain = 0;
    bool already_in_position= false;
    bool new_loop = false;

    std::vector<int> zero_cost_abs;
    for (auto active_i_pair:i_order){
        int active_i = active_i_pair.first;
        int no_gain = 0;
        pExt = bestPExt;
        while(!extension_queues[active_i].empty() && no_gain <= GlobalData::no_gain_counter_threshold_interleaving){
            if (!already_in_position && !new_loop){
                if (last_gain > highest_gain){
                    highest_gain = last_gain;
                    bestPExt = pExt;
                    abs_ext_best_i = abs_ext_list.size();
                }else{
                    no_gain++;
                    zero_cost_abs.push_back(abs_ext_list.size());
                }
                abs_ext_list.push_back(abs_ext);
            }
            new_loop = false;

            auto next = extension_queues[active_i].top();
            extension_queues[active_i].pop();
            auto after_i = next.gap_distribution_of;
            auto event = next.relative_to;

            //check if there is already en after_i event
            bool already_i_inExt = false;
            for (auto e:pExt.extensions){
                if (e.relativLocation == after_i){
                    already_i_inExt = true;
                    auto ext = pExt.getLocationExtension(after_i);
                    already_in_position = false;
                    if (event > db.getHighestEvent()){
                        for (auto abs:getAllAbstractionsOfEvent(ext.event,AbstractionExtension(),abs_ext_list)){
                            if(abs == event || abs == ext.event){
                                already_in_position = true;
                                break;
                            }
                        }
                    }else{
                        for (auto e:getAllChildrenOfEvent(ext.event, AbstractionExtension(), abs_ext_list, {}, true)){
                            if (e == event || e == ext.event){
                                already_in_position = true;
                                break;
                            }
                        }
                    }

                    //######################################
                    if (!already_in_position) abs_ext = extendPattern(pExt, after_i, event, true, abs_ext_list);
                }
            }
            if (!already_i_inExt){
                abs_ext = extendPattern(pExt, after_i, event, false, abs_ext_list);
            }
            if (already_in_position) continue; //skip rest of loop if nothing was added
//            auto copy_abs_ext_list = abs_ext_list;
//            copy_abs_ext_list.push_back(abs_ext);
//            last_gain = computeGain(*copy_p,copy_abs_ext_list);
            last_gain = estimatePatternGain(pExt, counts_after_i, abs_ext_list, abs_ext, zero_cost_abs);

        }
        if (last_gain > highest_gain){
            highest_gain = last_gain;
            bestPExt = pExt;
            abs_ext_best_i = abs_ext_list.size();
        }
        abs_ext_list.push_back(abs_ext);
        abs_ext_list.resize(abs_ext_best_i+1); //remove all abs_ext added after best ext.
        new_loop = true;
    }


    return ExtendCandidateReturn(highest_gain,&pattern, bestPExt.buildPattern(), abs_ext_list, bestPExt);
}

std::vector<int> Model::findMatchingAbstraction(std::set<int> children) {
    std::vector<int> matches;
    for (auto id_abstraction:abstractions){
        bool match = true;
        for (auto c:id_abstraction.second.getChildren()){
            if (children.find(c) == children.end()){
                match = false;
                break;
            }
        }
        if (match){
            matches.push_back(id_abstraction.first);
        }
    }
    return matches;
}

std::map<int, Abstraction> Model::getAbstractions() {
    return abstractions;
}


void Model::removePatternFromSelection(Pattern *p) {
    removePatternFromSelection(p->getId());
}

void Model::removePatternFromSelection(int patternId) {
    selection.erase(patternId);
}

int Model::positionToRelativ(int pattern_length, int position){
    if (position == pattern_length) {
        return 1;
    }else if (position == 0){
        return -1;
    }
    throw std::runtime_error("case not implemented/ not possible ");
}

void Model::update_pq_insert_positions(std::map<int, int> &pq_insert_positions, int new_position) {
    for (auto& i :pq_insert_positions){
        if (new_position > 0 && i.second >= new_position){
            i.second += 1;
        }
        if (new_position < 0 && i.second <= new_position){
            i.second -= 1;
        }
    }
}

void print_pq_pos(std::map<int,int> pq_insert_positions){
    for (auto e:pq_insert_positions){
        std::cout << e.first <<  " - " << e.second << std::endl;
    }
}

void print_pq(pq_gap_dis_ext pq){
    while(!pq.empty()){
        auto next = pq.top();
        pq.pop();
        next.print();
    }
}

ExtendCandidateReturn Model::extendCandidate(Candidate &candidate) {
    int best_candidate = 0;
    double best_estimate = 0.0;
    pq_gap_dis_ext extension_queue;
    std::set<int> already_added;
    std::map<int,int> pq_insert_positions;

    PatternExtension patternExtension = PatternExtension(candidate.getPattern());

    auto candidateExtensions = candidate.getExtensions();
    for (auto dst:candidateExtensions){
        double estimate = estimateGapFrequGain(candidate.getPattern(),dst.second);
        if (best_estimate < estimate) {
            best_candidate = dst.first;
            best_estimate = estimate;
        }
    }


    std::vector<AbstractionExtension> abstraction_extension_list;
    PatternExtension current_best = patternExtension;
    PatternExtension best_extension = patternExtension;


    auto new_position = positionToRelativ(candidate.getPattern().getEvents().size(),candidate.getPosition());
    auto bestExtAbsAction = extendPattern(best_extension, new_position, best_candidate, false, abstraction_extension_list);
    already_added.insert(best_candidate);

    double gain = estimatePatternGain(best_extension, candidateExtensions, abstraction_extension_list, bestExtAbsAction);






    double highest_gain = 0.0;
    addAllGapDistributionToPQ(extension_queue,candidate,best_candidate,already_added);
    pq_insert_positions[best_candidate] = new_position;

    bool skipped_loop = false;
    while ((gain - GlobalData::min_bit_gain_threshold > highest_gain || skipped_loop) && best_extension.getPatternSize() <= GlobalData::max_pattern_extension_factor*candidate.getPattern().getEvents().size() && !extension_queue.empty()){
        highest_gain = gain;

        current_best = best_extension;
        if (!skipped_loop){
            abstraction_extension_list.push_back(bestExtAbsAction);
        }
        PQEntryGapDistributionExtension next = extension_queue.top();

//        print_pq(extension_queue);

        extension_queue.pop();
        skipped_loop = false;
        if (already_added.find(next.relative_to) != already_added.end()){
            skipped_loop = true;
            continue;
        }
        // create 2 copys (c1 and c2) of pattern one for mean (left or right) extension
        PatternExtension c1 = current_best;
        PatternExtension c2 = current_best;
        // find relativ position of gap_distribution of

//        print_pq_pos(pq_insert_positions);
        int insert_position = pq_insert_positions[next.gap_distribution_of];

        // one for vertical extension
        int splice_in_position = insert_position;
        if (splice_in_position > 0 && candidate.getExtensions()[next.relative_to].getMStat() > candidate.getExtensions()[next.gap_distribution_of].getMStat()) { // if mean of extension is smaller
            splice_in_position++; // c1 =  splice new element into left side
        }else if (splice_in_position < 0 && candidate.getExtensions()[next.relative_to].getMStat() < candidate.getExtensions()[next.gap_distribution_of].getMStat()){
            splice_in_position--;
        }
        auto c1AbsExt = extendPattern(c1, splice_in_position, next.relative_to, false, abstraction_extension_list);


        auto can_ext = candidate.getExtensions();


        double c1_gain = estimatePatternGain(c1, candidateExtensions, abstraction_extension_list,c1AbsExt);

        // check if event already at location!
        auto ext = c2.getLocationExtension(insert_position);
        bool already_in_position = false;
        if (next.relative_to > db.getHighestEvent()){
            for (auto abs:getAllAbstractionsOfEvent(ext.event,AbstractionExtension(),abstraction_extension_list)){
                 if(abs == next.relative_to || abs == ext.event){
                     already_in_position = true;
                     break;
                 }
            }
        }else{
            for (auto e:getAllChildrenOfEvent(ext.event, AbstractionExtension(), abstraction_extension_list, {}, true)){
                if (e == next.relative_to || e == ext.event){
                    already_in_position = true;
                    break;
                }
            }
        }
        if (next.distance < GlobalData::divergence_threshold && !already_in_position){
            auto c2AbsExt = extendPattern(c2, insert_position, next.relative_to, true, abstraction_extension_list);


            double c2_gain = estimatePatternGain(c2, candidateExtensions, abstraction_extension_list, c2AbsExt);

            if (c1_gain > c2_gain){
                best_extension = c1;
                bestExtAbsAction = c1AbsExt;
                gain = c1_gain;
                update_pq_insert_positions(pq_insert_positions,splice_in_position);
                pq_insert_positions[next.relative_to] =  splice_in_position;

            }else{
                best_extension = c2;
                bestExtAbsAction = c2AbsExt;
                gain = c2_gain;
                pq_insert_positions[next.relative_to] =  insert_position;
            }

        }else{


            best_extension = c1;
            bestExtAbsAction = c1AbsExt;
            gain = c1_gain;
            update_pq_insert_positions(pq_insert_positions,splice_in_position);
            pq_insert_positions[next.relative_to] =  splice_in_position;

        }
        already_added.insert(next.relative_to);
        addAllGapDistributionToPQ(extension_queue, candidate, next.relative_to, already_added);

    }




    return ExtendCandidateReturn(highest_gain, &candidate.getPattern(), current_best.buildPattern(), abstraction_extension_list, current_best);
    // find closes match to any exntion
    // test extension added to abstraction and test for


    // at this point we are not done and we should test teh second best freu gap estimate, out of all non already added candidate extensions
}

double Model::estimateGapFrequGain(Pattern &p, GapDistribution& distribution) {
    if (distribution.getTotalCounts() == 0){
        return 0;
    }
    return (double) (double)distribution.getTotalCounts()/(double)abs(distribution.getMStat());

}

ReturnGainEstAbsExt Model::estimateGainFromAbsExt(std::vector<AbstractionExtension> prev_abs_ext,
                                     AbstractionExtension &abstractionExtension, std::vector<int> zero_cost_abs_ext) {
    int added_abstractions = 0;
    double delta_in_bits = 0;
    auto it = prev_abs_ext.begin();
    bool last_added = false;
    int i =0;
    int j =0;

    UniqueValueCounter<int> combined_pattern_frequ_updates;

    while (it != prev_abs_ext.end() || !last_added){
        if (zero_cost_abs_ext.size() > 0 && zero_cost_abs_ext[j] == i){
            j++;
            i++;
            continue;
        }

        AbstractionExtension abs;
        if (it == prev_abs_ext.end()){
            abs = abstractionExtension;
            last_added = true;
        }else{
            abs = *it;
        }
        auto abs_in_abs = 0;
        for (auto e:abs.events){
            if (e > db.getHighestEvent()) ++abs_in_abs;
        }

        if (abs.action == ADD_NEW){
            //extra bits required
            int i = abstractions.size() + added_abstractions;
            if (i > 0){
                delta_in_bits +=log2(i);
            }


            delta_in_bits +=logBinom(i,abs_in_abs);
            auto m = abs.events.size() - abs_in_abs;
            auto m_possible_children = db.getHighestEvent() - getAllChildrenOfAbsExt(abs,prev_abs_ext).size()+ m;

            delta_in_bits += log2(m_possible_children);
            delta_in_bits += logBinom(m_possible_children, m);
            ++added_abstractions;
        }else if (abs.action == EXTEND){
            auto prevCounts  = findExistingAbsinAbs(abs.id,prev_abs_ext,it);// number of abs_abs before ext
            auto old_indirect = prevCounts.all_children - prevCounts.direct_children;
            auto new_indirect = getAllChildrenOfAbsExt(abs,prev_abs_ext).size() - abs.events.size() - abs_in_abs - prevCounts.direct_children;


            auto old_possible_child = db.getHighestEvent() - old_indirect; // + prevCounts.direct_abstractions;
            auto new_possible_child = db.getHighestEvent() - new_indirect; //getAllChildrenOfAbsExt(abs,prev_abs_ext).size() + (prevCounts.direct_abstractions + abs_in_abs);
            if (abs_in_abs >0){ // if number of abs increases
                //problem we don't know i of abs / If not alreayd added does not have an i / also i of abs keet changing
                // => assume i to be |A| (number of abs) + added_abstractions
                int i = abstractions.size() + added_abstractions;
                int k = prevCounts.direct_abstractions;
                delta_in_bits -= logBinom(i,k);
                delta_in_bits += logBinom(i, k+abs_in_abs);


                delta_in_bits -= log2(old_possible_child);
                delta_in_bits += log2(new_possible_child);
            }

            delta_in_bits -= logBinom(old_possible_child,prevCounts.direct_children);
            delta_in_bits += logBinom(new_possible_child, prevCounts.direct_children + abs.events.size() - abs_in_abs);

            for (auto p_diff:abs.frequUpdates.getCounts()){
                combined_pattern_frequ_updates.incCount(p_diff.first, p_diff.second);
            }
        }
        if (last_added){
            break;
        }else{
            it++;
            i++;
        }
    }
    return ReturnGainEstAbsExt(delta_in_bits, combined_pattern_frequ_updates, added_abstractions);
}

/**
 * What would the encoding length be if the pattern added to selection
 * @param pattern
 * @return
 */
double Model::estimatePatternGain(PatternExtension pattern, extension_list counts) {

    auto emptyAbsExt = AbstractionExtension();
    return estimatePatternGain(pattern,counts,{},emptyAbsExt);

}


double Model::estimatePatternGain(Pattern pattern, int pattern_frequ, UniqueValueCounter<int> frequ_updates) {
    double delta_in_bits = 0;
    auto p_size = pattern.getEvents().size();
    delta_in_bits += universal_integer_encoding(p_size);
    delta_in_bits += p_size * -log2(1.0/(db.getHighestEvent()+abstractions.size()+1));

    delta_in_bits += estimate_delta_updated_frequ(pattern_frequ, frequ_updates);
    return delta_in_bits;
}


ReturnfindExistingAbsinAbs Model::findExistingAbsinAbs(int id, std::vector<AbstractionExtension>& prev_added, std::vector<AbstractionExtension>::iterator& limit)  {
    auto it = prev_added.begin();
    bool in_prev = false;
    std::set<int> local_abstractions;
    std::set<int> direct_children;
    std::set<int> all_children;
    std::vector<AbstractionExtension>::iterator last_id_usage;
    while (it != limit){
        if (it->id == id){
            if (it->action == ADD_NEW)in_prev = true;
            for (int child :it->events){
                if (child> db.getHighestEvent()){
                    local_abstractions.insert(child);
                }else{
                    direct_children.insert(child);
                    all_children.insert(child);
                }
            }
            last_id_usage = it;
        }
        it++;
    }
    if (!in_prev){
        auto children = abstractions.find(id)->second.getChildren();
        for (int child:children){
            if (child > db.getHighestEvent()){
                local_abstractions.insert(child);
            }else{
                direct_children.insert(child);
                all_children.insert(child);
            }
        }
        auto ac = getAllChildrenOfEvent(id);
        all_children.insert(ac.begin(),ac.end());
    }else{
        auto ac = getAllChildrenOfAbsExt(*last_id_usage, prev_added);
        all_children.insert(ac.begin(),ac.end());
    }
    return ReturnfindExistingAbsinAbs(local_abstractions.size(), direct_children.size(), all_children.size());
}



double Model::estimatePatternGain(PatternExtension &pattern, extension_list  counts, std::vector<AbstractionExtension> prev_abs_ext,AbstractionExtension &abstractionExtension){

    auto return_est_gain_abs = estimateGainFromAbsExt(prev_abs_ext, abstractionExtension);
    auto delta_in_bits = return_est_gain_abs.delta_in_bits;
    auto combined_pattern_frequ_updates = return_est_gain_abs.combined_pattern_frequ_updates;
    auto added_abstractions = return_est_gain_abs.added_abstractions;

//    combined_pattern_frequ_updates.print();

    // cost of all applying all abstractionExtension to model computed at this point



    //================

    //NEW Pattern frequ

    int new_pattern_frequ = pattern.pattern.getPatternUsageCounts();

    for (auto e:pattern.extensions){
        int max_frequ_e = 0;
        for (auto child:getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts.find(child);
            if (c_it != counts.end()){
                max_frequ_e +=  c_it->second.getTotalCounts();
            }
        }
        new_pattern_frequ = std::min(new_pattern_frequ,max_frequ_e);
    }


    //================

    std::vector<std::vector<int>> specificationFrequencies;

    combined_pattern_frequ_updates.incCount(pattern.pattern.getId(),-new_pattern_frequ);
    for (auto e:pattern.extensions){
        std::vector<int> e_spec_frequ;
        auto e_sum = 0;
        for (auto child:getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts.find(child);
            if (c_it != counts.end()){
                e_sum += c_it->second.getTotalCounts();
            }
        }
        for (auto child:getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts.find(child);
            if (c_it != counts.end()){
                auto diff = (c_it->second.getTotalCounts()/(double)e_sum) * new_pattern_frequ;
                combined_pattern_frequ_updates.incCount(child,-diff);
                e_spec_frequ.push_back(diff);

            }
        }
        specificationFrequencies.push_back(e_spec_frequ);
    }
//    combined_pattern_frequ_updates.print();

    double estimated_spec_cost = 0;
    std::vector<int> empty_vec;
    for (auto e_spec_frequ:specificationFrequencies){
        if (e_spec_frequ.size() > 1){
            estimated_spec_cost += estimate_specification_delta(empty_vec, e_spec_frequ);
        }
    }

    estimated_spec_cost += estimate_spec_delta_for_patterns(combined_pattern_frequ_updates);


    delta_in_bits += estimate_delta_updated_frequ(new_pattern_frequ, combined_pattern_frequ_updates);
    delta_in_bits += estimated_spec_cost;
    delta_in_bits -= ct_gain_by_zero_frequ(combined_pattern_frequ_updates);

    if (std::isinf(delta_in_bits) || std::isnan(delta_in_bits)){
        std::cout << delta_in_bits << std::endl;
        throw std::runtime_error("Estimate is nan or inf!");
    }



    return -delta_in_bits;

}



double
Model::estimatePatternGain(PatternExtension &pattern, std::vector<std::map<int, GapDistribution>> &counts_after_i,
                           std::vector<AbstractionExtension> prev_abs_ext, AbstractionExtension &abstractionExtension,
                           std::vector<int> zero_cost_abs) {
    auto return_est_gain_abs = estimateGainFromAbsExt(prev_abs_ext, abstractionExtension, zero_cost_abs);
    auto delta_in_bits = return_est_gain_abs.delta_in_bits;
    auto combined_pattern_frequ_updates = return_est_gain_abs.combined_pattern_frequ_updates;
    auto added_abstractions = return_est_gain_abs.added_abstractions;


    int new_pattern_frequ = pattern.pattern.getPatternUsageCounts();
    for (auto e:pattern.extensions){
        int max_frequ_e = 0;
        for (auto child: getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts_after_i[e.relativLocation].find(child);
            if (c_it != counts_after_i[e.relativLocation].end()){
                max_frequ_e += c_it->second.getTotalCounts();
            }
        }
        new_pattern_frequ = std::min(new_pattern_frequ, max_frequ_e);
    }

    combined_pattern_frequ_updates.incCount(pattern.pattern.getId(), -new_pattern_frequ);


    std::vector<std::vector<int>> specificationFrequencies;
    for (auto e:pattern.extensions){
        std::vector<int> e_spec_frequ;
        auto e_sum = 0;
        for (auto child:getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts_after_i[e.relativLocation].find(child);
            if (c_it != counts_after_i[e.relativLocation].end()){
                e_sum += c_it->second.getTotalCounts();
            }
        }
        for (auto child:getAllChildrenOfEvent(e.event, abstractionExtension, prev_abs_ext, {})){
            auto c_it = counts_after_i[e.relativLocation].find(child);
            if (c_it != counts_after_i[e.relativLocation].end()){
                auto diff = (c_it->second.getTotalCounts()/(double)e_sum) * new_pattern_frequ;
                combined_pattern_frequ_updates.incCount(child, -diff);
                e_spec_frequ.push_back(diff);
            }
        }
        specificationFrequencies.push_back(e_spec_frequ);

    }

    double estimated_spec_cost = 0;
    std::vector<int> empty_vec;
    for (auto e_spec_frequ:specificationFrequencies){
        if (e_spec_frequ.size() > 1){
            estimated_spec_cost += estimate_specification_delta(empty_vec, e_spec_frequ);
        }
    }

    estimated_spec_cost += estimate_spec_delta_for_patterns(combined_pattern_frequ_updates);


    delta_in_bits += estimate_delta_updated_frequ(new_pattern_frequ, combined_pattern_frequ_updates);
    delta_in_bits += estimated_spec_cost;
    delta_in_bits -= ct_gain_by_zero_frequ(combined_pattern_frequ_updates);

    if (std::isinf(delta_in_bits) || std::isnan(delta_in_bits)){
        std::cout << delta_in_bits << std::endl;
        throw std::runtime_error("Estimate is nan or inf!");
    }

    return -delta_in_bits;



}

void Model::addAllGapDistributionToPQ(pq_gap_dis_ext &pq, Candidate &candidate,
                                      int for_element ,std::set<int> already_added) {


    GapDistribution base_case = candidate.getExtensions()[for_element];
    for (auto id_gapDst:candidate.getExtensions()){
        if(already_added.find(id_gapDst.first) == already_added.end() && id_gapDst.first != for_element){
            double div = base_case.divergence_frqu_adjsted(id_gapDst.second);
            auto pq_entry = PQEntryGapDistributionExtension(for_element,id_gapDst.first, div);
            pq.push(pq_entry);
        }
    }
}

double Model::computeGain(Pattern &pattern, std::vector<AbstractionExtension> prev_abs_ext, bool ignore_spec_cost) {
    WindowSet cover_save = cover;
    if (ignore_spec_cost) computeLength(ignore_spec_cost);
    auto prev_length = encoded_length;
    applyAbstractionExtensionList(prev_abs_ext);
    addPatternToSelection(&pattern, false);
    auto pattern_windows = findWindows(pattern, GlobalData::window_budget);
    pattern_windows.sort();
    cover = greedyCover(pattern_windows);
    recomputeUsages();
    computeLength(ignore_spec_cost);
    pattern.copyLocations();
    double return_value = prev_length - encoded_length;
    // reset cover to previous version
    removePatternFromSelection(&pattern);
    undoAbstractionExtensionList(prev_abs_ext);
    cover = cover_save;
    recomputeUsages();
    computeLength();
    return return_value;
}

double Model::computeGain(Pattern &pattern, std::vector<AbstractionExtension> prev_abs_ext, Pattern &source_pattern){
    WindowSet cover_save = cover;
    auto prev_length = encoded_length;
    applyAbstractionExtensionList(prev_abs_ext);
    addPatternToSelection(&pattern, false);
    auto pattern_windows = findWindows(pattern, GlobalData::window_budget);
    pattern_windows.sort();
    removePatternFromSelection(&source_pattern);
    availableWindowSet.removeWindowsOfPattern(source_pattern.getId());
    cover = greedyCover(pattern_windows);
    recomputeUsages();
    computeLength();
    for (auto w:cover){
        std::cout << w.getPattern()->getId() << ", ";
    }
    std::cout << std::endl;
    pattern.copyLocations();
    double return_value = prev_length - encoded_length;
    // reset cover to previous version
    removePatternFromSelection(&pattern);
    undoAbstractionExtensionList(prev_abs_ext);
    addPatternToSelection(&source_pattern, false);
    auto source_pattern_windows = findWindows(source_pattern,GlobalData::window_budget);
    availableWindowSet.addToWindowSet(source_pattern_windows);
    cover = cover_save;
    recomputeUsages();
    computeLength();
    return return_value;
}

double Model::computeGainAbsExt(std::vector<AbstractionExtension> abs_ext_list, Pattern &pattern) {
    auto prev_length = encoded_length;
    WindowSet cover_save = cover;
    applyAbstractionExtensionList(abs_ext_list);
    auto windows = findWindows(pattern, GlobalData::window_budget);
    windows.sort();
    cover = greedyCover(windows);
    recomputeUsages();
    computeLength();
    double return_value = prev_length - encoded_length;
    undoAbstractionExtensionList(abs_ext_list);
    cover = cover_save;
    recomputeUsages();
    computeLength();
    return return_value;
}


/*
 * Main search loop
 */
void Model::search(bool in_search_pruning, bool prune_with_recover, bool print_steps) {
    double highest_gain = 1.0;
    std::set<std::set<int>> seen_open;
    open = this->getSelection(); //init open
    bool gain = true;
    std::set<int> removedPatterns;
    while (gain) {
        gain = false;
        std::priority_queue<ExtendCandidateReturn> extensions;
        for (auto id_pattern:open) {
            Pattern *pattern = id_pattern.second;

            // refine step
            auto gain_extendedPatterns = refinePattern(*pattern);

            for (auto gain_extendedPattern:gain_extendedPatterns){
                if (0.0 < gain_extendedPattern.estimatedGain) {
                    if (print_steps){
                        std::cout << "Refining pattern from: " << std::endl;
                        id_pattern.second->print();
                        std::cout << "TO" << std::endl;
                        gain_extendedPattern.best_extension->print();
                        std::cout << std::endl;
                    }
                    extensions.push(gain_extendedPattern);
                }
            }

        }
        open.clear();

        int no_gain_counter = 0;
//        std::map<int,int> source_patterns_lookup;
//        std::map<int,std::set<int>> patterns_from_source;
        while(!extensions.empty() && no_gain_counter < 100){

            auto ext_cand_return = extensions.top();
            extensions.pop();
            ext_cand_return.best_extension->setId(++patternCounter);
            auto no_abs_version = cutOutAbstractions(ext_cand_return);
            double realGain  = computeGain(*ext_cand_return.best_extension, ext_cand_return.abstractionExtensions, true);
            double realGainNoAbs = computeGain(*no_abs_version.best_extension, no_abs_version.abstractionExtensions,
                                               true);

            if (realGainNoAbs > realGain){
                realGain = realGainNoAbs;
                ext_cand_return = no_abs_version;
            }

            if (print_steps){
                std::cout << "Testing pattern for addition: " << std::endl;
                ext_cand_return.best_extension->print();
                std::cout << "Abstraction Extensions" << std::endl;
                for (auto a:ext_cand_return.abstractionExtensions){
                    a.print();
                }
                std::cout << "Gain by Pattern: " << realGain << std::endl << std::endl;
            }



            if (realGain > 0.0){

                WindowSet cover_pre_add = cover;

                ext_cand_return = estimateGainOnRealCounts(ext_cand_return, realGain);

                no_gain_counter = 0;
                applyAbstractionExtensionList(ext_cand_return.abstractionExtensions);
                computePattenStructure(ext_cand_return.best_extension);
                addPatternToSelection(ext_cand_return.best_extension, false);

                if (removedPatterns.find(ext_cand_return.source_pattern->getId()) == removedPatterns.end()){
                    open[ext_cand_return.source_pattern->getId()] = ext_cand_return.source_pattern;
//                    source_patterns_lookup[ext_cand_return.best_extension->getId()] = ext_cand_return.source_pattern->getId();
//                    patterns_from_source[ext_cand_return.source_pattern->getId()].insert(ext_cand_return.best_extension->getId());
                }
                WindowSet extension_window_set = findWindows(*ext_cand_return.best_extension, GlobalData::window_budget);
                extension_window_set.sort();
                availableWindowSet.addToWindowSet(extension_window_set);
                cover = greedyCover();
                recomputeUsages();
                computeLength();
                gain = true;


                auto pExt = refinePatternInterleaving(*ext_cand_return.best_extension);

                if (print_steps){
                    std::cout << "Interleaved Extended pattern: " << std::endl;
                    pExt.best_extension->print();
                    std::cout << "Abstraction Extensions: " << std::endl;
                    for (auto a:pExt.abstractionExtensions){
                        a.print();
                    }
                }

                WindowSet cover_save = cover;
                removePatternFromSelection(ext_cand_return.best_extension);
                availableWindowSet.removeWindowsOfPattern(ext_cand_return.best_extension->getId());
                cover = cover_pre_add;
                recomputeUsages();
                computeLength();

                double computed_gain = 0;
                if (pExt.estimatedGain > 0){
                    computed_gain =  computeGain(*pExt.best_extension, pExt.abstractionExtensions);
                }
                if (pExt.estimatedGain > 0 && (computed_gain > 0)){
                    if (print_steps) std::cout << "Adding to selection " << std::endl;
                    applyAbstractionExtensionList(pExt.abstractionExtensions);
                    computePattenStructure(pExt.best_extension);
                    addPatternToSelection(pExt.best_extension, false);
                    open[pExt.best_extension->getId()] = pExt.best_extension;
                    WindowSet  extension_window_set = findWindows(*pExt.best_extension, GlobalData::window_budget);
                    extension_window_set.sort();
                    availableWindowSet.addToWindowSet(extension_window_set);
                    cover = greedyCover();
                    recomputeUsages();
                    computeLength();
                }else{
                    addPatternToSelection(ext_cand_return.best_extension, false);
                    availableWindowSet.addToWindowSet(extension_window_set);
                    cover = cover_save;
                    recomputeUsages();
                    computeLength();

                    if (print_steps) std::cout << "Not added to selection" << std::endl;
                    open[ext_cand_return.best_extension->getId()] = ext_cand_return.best_extension;
                }
            }else{
                no_gain_counter++;
            }
        }
        if (print_steps) printSelection(false);
        while (true){
            auto prunedStuff = mergePatternsAndFlattenAbs();
            removedPatterns.insert(prunedStuff.first.begin(),prunedStuff.first.end());
            if (in_search_pruning){
                std::vector<int> prunedPatterns;
                if (prune_with_recover){
                    prunedPatterns = pruneWithReCover();
                }else{
                    prunedPatterns= prunePatterns();
                }
                removedPatterns.insert(prunedPatterns.begin(), prunedPatterns.end());
                auto morePrunedPatterns = pruneWithReCover();
                removedPatterns.insert( morePrunedPatterns.begin(), morePrunedPatterns.end());
            }
            for (auto p_id:removedPatterns){
                open.erase(p_id);
//                if (source_patterns_lookup.find(p_id) != source_patterns_lookup.end()){
//                    patterns_from_source.find(source_patterns_lookup[p_id])->second.erase(p_id);
//                    if (patterns_from_source.find(source_patterns_lookup[p_id])->second.empty()){
//                        open.erase(source_patterns_lookup[p_id]);
//                    }
//                }
            }
            for (auto replace_abs:prunedStuff.second){
                if (replace_abs.second == -1) continue;
                replaceSymbolInOpen(replace_abs.first,replace_abs.second);
            }
            if (prunedStuff.first.empty() && prunedStuff.second.empty()){
                break;
            }
        }
//        source_patterns_lookup.clear();
//        patterns_from_source.clear();

        std::set<int> last_open;
        for (auto p:open){
            last_open.insert(p.first);
        }
        if (seen_open.find(last_open) == seen_open.end()){
            seen_open.insert(last_open);
        } else{
            break;
        }
//        extendAbstractions();
    }
    if (!in_search_pruning){
        prunePatterns();
    }
    removeUnusedAbstractions();
    if (prune_with_recover){
        pruneWithReCover();
    }
    removeUnusedAbstractions();
    recomputeUsages();
    computeLength();
}

void Model::printAbstractions() { 
    for (auto ab:abstractions){
        std::cout << "Id: " << ab.second.getId() << " - ";
        auto children = ab.second.getChildren();
        std::sort(children.begin(),children.end());
        for (auto child:children){
            std::cout << child << " ";
        }
        std::cout << std::endl;
    }
}

void Model::applyAbstractionExtensionList(std::vector<AbstractionExtension> extensions) {
    for (auto e:extensions){
        applyAbstractionExtension(e);
    }
}

void Model::applyAbstractionExtension(AbstractionExtension extension) {
    switch (extension.action) {
        case DO_NOTHING:
            break;
        case EXTEND:
            extendAbstraction(extension.id,extension.events);
            break;
        case ADD_NEW:
            addAbstraction(Abstraction(extension.id, extension.events), false);
            break;
    }
}

void Model::undoAbstractionExtension(AbstractionExtension extension) {
    switch (extension.action) {
        case DO_NOTHING:
            break;
        case EXTEND:
            abstractions.find(extension.id)->second.removeChildrenFromAbstraction(extension.events);
            break;
        case ADD_NEW:
            removeAbstraction(extension.id);
            break;
    }
}

void Model::undoAbstractionExtensionList(std::vector<AbstractionExtension> extension) {
    auto e_it = extension.rbegin();
    while (e_it != extension.rend()){
        undoAbstractionExtension(*e_it);
        e_it++;
    }
}

const std::map<int, int> &Model::getUsageCounts() {
    return pattern_singleton_usage;
}

bool Model::subSet(std::vector<int>& sub, std::vector<int>& par){
    for (int se:sub){ 
        bool found = false;
        for (int pe:par){
            if (se == pe){
                found = true;
                break;
            }
        }
        if (!found){
            return false;
        }
    }
    return true;
}

int Model::findSmallestSuperSetAbstraction(AbstractionExtension ext, std::vector<AbstractionExtension> abstractionExtensionList) {
    if (ext.action != ADD_NEW){
        throw std::runtime_error("Has to be new abstraction");
    }
    int smallest_set = -1;
    int smallest_set_size = INT32_MAX;

    auto ext_children = getAllChildrenOfAbsExt(ext, abstractionExtensionList);
    for (auto i_ai:abstractions){
        std::vector<int> abs_children = getAllChildrenOfAbstraction(i_ai.second);
        if (subSet(ext_children,abs_children) && abs_children.size() < smallest_set_size){
            smallest_set = i_ai.first;
            smallest_set_size = abs_children.size();
        }
    }
    return smallest_set;
}


std::vector<ExtendCandidateReturn> Model::refineCandidate(Candidate& pattern_candidate){
    auto split_cand = splitSingletonAndPatterns(pattern_candidate);
    auto extensions = patternCandidatesToPattern(split_cand.second);
    extendCandidateWithAbstractions(split_cand.first);
    ExtendCandidateReturn singletonExtension =  extendCandidate(split_cand.first);
    extensions.push_back(singletonExtension);
    return extensions;
}

ExtendCandidateReturn Model::refinePatternInterleaving(Pattern& p) {
    auto counts_after_i = extendInterleaving(p);

    for (auto& c:counts_after_i){
        auto dummyCandidate = Candidate(p,0,c);
        extendCandidateWithAbstractions(dummyCandidate);
        c = dummyCandidate.getExtensions();
    }

    auto pExt = extendPattern(p,counts_after_i);
    pExt.best_extension->setId(++patternCounter);
    return pExt;

}

std::vector<ExtendCandidateReturn> Model::refinePattern(Pattern& p) {
    Candidate pattern_candidate_f = extendByFollowing(p);
    auto extensions = refineCandidate(pattern_candidate_f);
    Candidate pattern_candidate_p = extendByProceeding(p);
    auto p_extensions = refineCandidate(pattern_candidate_p);
    extensions.insert(extensions.end(),p_extensions.begin(), p_extensions.end());
    return extensions;
}

std::pair<Candidate, Candidate> Model::splitSingletonAndPatterns(Candidate& candidate) {
    extension_list singleton_extensions;
    extension_list pattern_extension;
    for (auto ext:candidate.getExtensions()){
        if (ext.first <= db.getHighestEvent()){
            singleton_extensions[ext.first] = ext.second;
        }else{
            pattern_extension[ext.first] = ext.second;
        }
    }
    auto singCand = Candidate(candidate.getPattern(), candidate.getPosition(), singleton_extensions);
    auto pattCand = Candidate(candidate.getPattern(), candidate.getPosition(), pattern_extension);
    return std::pair<Candidate,Candidate>(singCand, pattCand);
}

std::vector<ExtendCandidateReturn> Model::patternCandidatesToPattern(Candidate candidate) {
    auto ext_list = candidate.getExtensions();
    auto ext_as_array = std::vector<std::pair<int,GapDistribution>>(ext_list.begin(), ext_list.end());
    struct {
        bool operator()(std::pair<int,GapDistribution> lhs, std::pair<int,GapDistribution> rhs) const{
            return lhs.second.getTotalCounts() > rhs.second.getTotalCounts();
        }
    } frequComp;
    std::sort(ext_as_array.begin(), ext_as_array.end(), frequComp);




    std::vector<ExtendCandidateReturn> new_patterns;
    for (auto next_ext: ext_as_array){


        Pattern* c = new Pattern(candidate.getPattern());




        c->extendWithPattern(*selection[next_ext.first],candidate.getPosition());


        auto new_pattern_frequ = next_ext.second.getTotalCounts();
        UniqueValueCounter<int> frequ_updates;
        frequ_updates.incCount(candidate.getPattern().getId(),-new_pattern_frequ); //old pattern usage decreses by new_pattern_frequ
        frequ_updates.incCount(next_ext.first, -new_pattern_frequ); // as well as the pattern used to extend
        double estimated_gain = estimatePatternGain(*c, new_pattern_frequ, frequ_updates);
        if (estimated_gain > 0.0){
            auto exCanRet  = ExtendCandidateReturn(estimated_gain, &candidate.getPattern(), c, {});
            new_patterns.push_back(exCanRet);
        }
    }
    return new_patterns;
}

bool similarEnough(GapDistribution lhs, GapDistribution rhs){
    return GlobalData::divergence_threshold > lhs.divergence(rhs)/std::min(lhs.getTotalCounts(),rhs.getTotalCounts());
//    return GlobalData::divergence_threshold > lhs.divergence(rhs);
}

void Model::extendCandidateWithAbstractions(Candidate& candidate) {
    auto ext_list = candidate.getExtensions();
    std::vector<int> absToAdd;
    for (auto abs:abstractions){
        auto all_children = getAllChildrenOfAbstraction(abs.second);
        bool added = false;
        for (auto c1:all_children){
            for (auto c2:all_children){
                if (c1 < c2 &&// to avoid testing (a,b) and (b,a) as well as (a,a)
                   ext_list.find(c1) != ext_list.end() && ext_list.find(c2) != ext_list.end() &&
                   similarEnough(ext_list[c1], ext_list[c2]))
                {
                    absToAdd.push_back(abs.first);
                    added = true;
                    break;
                }
            }
            if (added){
                break;
            }
        }
    }
    for (int id:absToAdd){
        auto newGapDst = GapDistribution();
        auto allChildren = getAllChildrenOfAbstraction(getAbstraction(id));
        for (auto child: allChildren){
            if (ext_list.find(child) != ext_list.end()){
                auto child_dst = ext_list[child];
                for (auto delta:child_dst.getAllDeltas()){
                    newGapDst.incDeltaFrequency(delta,child_dst.getDeltaFrequency(delta));
                }
            }
        }
        candidate.addExtensionToList(id,newGapDst);
    }
}



double Model::estimate_delta_updated_frequ(int new_pattern_frequ, UniqueValueCounter<int> frequ_updates) {
    double delta_in_bits = 0;
    int total_old_frequ = 0;
    for (auto p:selection){
        total_old_frequ += p.second->getPatternUsageCounts();
    }




    int total_new_frequ = total_old_frequ  + new_pattern_frequ;
    for (auto f:frequ_updates.getCounts()){
//        int frequency = std::min(f.second,0); // limit frequency to 0
        int frequency = f.second;
        auto old_frequ = selection[f.first]->getPatternUsageCounts();
        auto new_frequ = old_frequ + frequency; // apply diff to pattern frequ

        if (new_frequ > 0 && old_frequ > 0){ // we limit frequ to 0 i.e. no negative frequencies
            delta_in_bits += old_frequ*log2(old_frequ) - (new_frequ * log2(new_frequ));
            total_new_frequ += frequency;
        }
        else if (old_frequ > 0){
            delta_in_bits += old_frequ*log2(old_frequ);
            total_new_frequ -= old_frequ;
        } else if (new_frequ > 0){
            delta_in_bits -= new_frequ * log2(new_frequ);
            total_new_frequ += frequency;
        }

    }
    if (new_pattern_frequ > 0){
        delta_in_bits -= (new_pattern_frequ) *log2(new_pattern_frequ);
    }
    delta_in_bits += total_new_frequ * log2(total_new_frequ);
    delta_in_bits -= total_old_frequ * log2(total_old_frequ);

    return delta_in_bits;
}

double compute_spec_cost(std::vector<int>& frequ){
    double episilon = 0.5;
    double delta_in_bits = 0;
    int base_count = 0;
    int old_f_size = frequ.size();
    for (auto fi:frequ){
        int i = 0;
        while (i < fi){
            delta_in_bits -= log2((episilon + i)/(old_f_size*episilon + base_count));
            i++;
            base_count++;
        }
    }
    return delta_in_bits;
}

double Model::estimate_specification_delta(std::vector<int> &old_f, std::vector<int> &new_f) {
    return -compute_spec_cost(old_f) + compute_spec_cost(new_f);
}

double Model::estimate_spec_delta_for_patterns(UniqueValueCounter<int> combined_pattern_frequ_updates) {
    double estimated_spec_cost = 0;
    for (auto f_update:combined_pattern_frequ_updates.getCounts()){
        auto pattern = selection[f_update.first];
        auto p_events = pattern->getEvents();
        auto p_size = p_events.size();
        int old_f = pattern->getPatternUsageCounts();
        int new_f = old_f + f_update.second;

        for (int i =0; i < p_size; i++){
            auto e = p_events[i];
            if (e > db.getHighestEvent()){
                std::vector<int> old_f_e;
                std::vector<int> new_f_e;
                for (auto child_usage:pattern->getUnconditionalUsageCounts(i).getCounts()){
                    old_f_e.push_back(child_usage.second);
                    new_f_e.push_back((new_f/(double)old_f) * child_usage.second);
                }
                estimated_spec_cost += estimate_specification_delta(old_f_e, new_f_e);
            }
        }

    }
    return estimated_spec_cost;
}

double Model::ct_gain_by_zero_frequ(UniqueValueCounter<int> combined_pattern_frequ_updates){
    double saved_ct_cost =0;
    for (auto f_update:combined_pattern_frequ_updates.getCounts()){
        auto pattern = selection[f_update.first];
        saved_ct_cost += computePatternLength(*pattern);
    }
    return saved_ct_cost;
}

void Model::printSelection(bool ignore_singletons) {
    std::cout << "Found Patterns: " << std::endl;
    for (auto p:getSelection()) {
        if (!ignore_singletons || p.first > db.getHighestEvent()){
            p.second->print();
            int usage_count = getUsageCounts().find(p.first)->second;
            std::cout << "Usage count: "  << usage_count << std::endl;
        }
    }

    std::cout << "Found Abstractions" << std::endl;
    printAbstractions();
}

std::string Model::getPatternOutputString(bool ignore_singletons,  bool incl_trans_p){
    std::string output;
    for (auto p:getSelection()){
        if (!ignore_singletons || p.first > db.getHighestEvent()){
            output += p.second->getOutputString();
            if (incl_trans_p){
                output += p.second->getTransPasString();
            }
            output += + "\n";
        }
    }
    return output;
}

std::string Model::getAbstractionOutputString(){
    std::vector<std::pair<int,Abstraction>> abs_as_vec;
    for (auto a:abstractions){
        abs_as_vec.push_back(a);
    }
    std::sort(abs_as_vec.begin(), abs_as_vec.end());
    std::string output;
    for (auto a:abs_as_vec){
        output += a.second.getOutputString();
    }
    return output;
}

//pruning of abs ext??
ExtendCandidateReturn Model::estimateGainOnRealCounts(ExtendCandidateReturn addition_candidate, double gain_to_beat) {
//    if (!addition_candidate.contains_pExt){
        int event_counter = 0;
        std::vector<int> event_ids;
        for (auto e:addition_candidate.best_extension->getEvents()){
            if (e > db.getHighestEvent()){
                for (auto a:addition_candidate.abstractionExtensions){
                    if (e == a.id){
                        // identified an abstraction change made to the pattern
//                        addition_candidate.best_extension->getCopyLocations()
                        event_ids.push_back(event_counter);
                        break;
                    }
                }
            }
            event_counter++;
        }
        std::vector<UniqueValueCounter<int>> cover_counts(event_counter);
        for (auto w:addition_candidate.best_extension->getCopyLocations()){
            auto positions = w.getPositions();
            for (auto e:event_ids){
                cover_counts[e].incCount(db.getEvent(positions[e]),1);
            }
        }


        // remove all from abs that do not occur in counts

        std::map<int,AbstractionExtension> newAbsExt;
        for (auto e:event_ids){
            int abs_id = addition_candidate.best_extension->getEvents()[e];
            bool is_new = false;
            for (auto a:addition_candidate.abstractionExtensions){
                if (a.id == abs_id && a.action == ADD_NEW){
                    is_new = true;
                }
            }
            AbstractionExtension a;
            if (is_new){
                a = AbstractionExtension(ADD_NEW, cover_counts[e].getKeys(), abs_id);
            }else{
                a = AbstractionExtension(EXTEND, cover_counts[e].getKeys(), abs_id);
            }
            newAbsExt[abs_id] = (a);
        }



        std::priority_queue<PQEntryRealCounts, std::vector<PQEntryRealCounts>, std::greater<PQEntryRealCounts>> pq;
        for (auto e:event_ids){
            for (auto c:cover_counts[e].getCounts()){
                PQEntryRealCounts pqEntry(e,c.first, c.second);
                pq.push(pqEntry);
            }
        }


        double best_gain = gain_to_beat;
        std::vector<AbstractionExtension> newAbsExtAsList;
        for (auto a:newAbsExt){
            newAbsExtAsList.push_back(a.second);
        }
        while (!pq.empty()){
            auto o = pq.top();
            pq.pop();
            int abs_id = addition_candidate.best_extension->getEvents()[o.pattern_e_index];
            AbstractionExtension rollback;
            Pattern* rollbackPattern;
            bool p_rollback = false;
            auto& a = newAbsExt[abs_id];

            for (auto e = a.events.begin(); e != a.events.end();){
                if (*e == o.event_id){
                    e = a.events.erase(e);
                    rollback = AbstractionExtension(EXTEND, {*e}, a.id);
                }else{
                    e++;
                }
            }
            if(a.events.empty()){
                rollback = AbstractionExtension(a.action,rollback.events, a.id);
                rollbackPattern = new Pattern(*addition_candidate.best_extension);
                p_rollback = true;
                int i =0;
                while (i < addition_candidate.best_extension->getEvents().size()){
                    auto pi = addition_candidate.best_extension->getEvents()[i];
                    if (pi == a.id){
                        addition_candidate.best_extension->removeEvent(i);
                    }else{
                        i++;
                    }
                }
                newAbsExt.erase(abs_id);
            }


            newAbsExtAsList.clear();
            for (auto a:newAbsExt){
                newAbsExtAsList.push_back(a.second);
            }
            double newGain = computeGain(*addition_candidate.best_extension, newAbsExtAsList);
            if (best_gain < newGain){
                best_gain = newGain;
                if (p_rollback) delete rollbackPattern;
            }else{
                if (p_rollback) addition_candidate.best_extension = rollbackPattern;
                newAbsExtAsList.push_back(rollback);
                break;
            }

        }
    addition_candidate.abstractionExtensions = newAbsExtAsList;
    return addition_candidate;
}
//// overlap of one is not detected, because irrelevant
//// returns largest continous overlap between two sequences (not efficient)
//std::vector<int> computeOverlap(std::vector<int> p1, std::vector<int> p2){
//    int L_p1 = 0;
//    std::vector<std::pair<int,int>> best_window = std::vector<std::pair<int,int>>(p1.size());
//    int best_i = -1;
//    while (L_p1 < p1.size()){
//        best_window[L_p1] = {0,0};
//        int L_p2 = 0;
//        int R_p2 = 0;
//        while (L_p2 < p2.size()){
//            while (p2[L_p2] != p1[L_p1] && L_p2 < p2.size()){
//                L_p2++;
//            }
//            R_p2 = L_p2 + 1;
//            while (p2[R_p2] == p1[L_p1 + R_p2 -L_p2] && R_p2 < p2.size()){
//                R_p2++;
//            }
//            R_p2--;
//            if (best_window[L_p1].second < R_p2 - L_p2){
//                best_window[L_p1] = {L_p2, R_p2 - L_p2};
//            }
//            L_p2 = R_p2 + 1;
//        }
//        if (best_window[L_p1].second > 0 && (best_i == -1 || best_window[best_i].second < best_window[L_p1].second)){
//            best_i = L_p1;
//        }
//        if (best_window[L_p1].second == 0){
//            L_p1++;
//        }else{
//            L_p1 = L_p1 + best_window[L_p1].second;
//        }
//    }
//    if (best_i != -1){
//        return {best_i,best_window[best_i].first,best_window[best_i].second};
//    }else{
//        return {best_i,-1,-1};
//    }
//}

std::vector<int> computeOverlap(std::vector<int> p1, std::vector<int> p2) {
    if (p1.size() != p2.size()){
        throw std::runtime_error("Pattern lengths don't match.");
    }
    std::vector<int> overlap;
    for (int i =0; i < p1.size(); i++){
        if (p1[i] == p2[i]){
            overlap.push_back(i);
        }
    }
    return overlap;
}

std::pair<std::vector<int>, std::vector<std::pair<int,int>>> Model::mergePatternsAndFlattenAbs() {

    std::vector<Pattern*> removeSet;
    std::set<int> removed_pattern_ids;
    for (auto p:selection){
        if (p.first > db.getHighestEvent()){
            if (getUsageCounts().find(p.first)->second <= 0){
                removeSet.push_back(p.second);
                removed_pattern_ids.insert(p.first);
            }
        }

    }
    for (auto p:removeSet){
        removePatternFromSelection(p);
        availableWindowSet.removeWindowsOfPattern(p->getId());
    }
    std::vector<std::pair<int,int>> abstractionRemoveSet;
    for (auto a:abstractions){
        auto all_children = getAllChildrenOfEvent(a.first);
        if (all_children.empty()){
            abstractionRemoveSet.emplace_back(a.first,-1);
        }else if (all_children.size() == 1){
            auto the_child = all_children[0];
            replaceSymbolInPatterns(a.first,the_child);
            for (auto a2:abstractions){
                auto abs = getAbstraction(a2.first);
                for (auto c:a2.second.getChildren()){
                    if (c == a.first){
                        abs.removeChildrenFromAbstraction({c});
                        abs.extendAbstraction(the_child);
                    }
                }
            }
            abstractionRemoveSet.emplace_back(a.first,-1);
        }
    }
    for (auto a:abstractionRemoveSet){
        removeAbstraction(a.first);
    }
    recomputeUsages();
    computeLength();

    pq_pattern_overlap overlap_queue;

    for (auto p1:selection){
        if (p1.first <= db.getHighestEvent()) continue;
        int best = -1;
        std::vector<int> best_overlap = std::vector<int>(3);
        int best_combined_usage  = 0;
        for (auto p2:selection){
            if (p2.first <= db.getHighestEvent()) continue;
            if (p1.first < p2.first){
                if (p1.second->getEvents().size() != p2.second->getEvents().size()) continue;
                auto overlap = computeOverlap(p1.second->getEvents(), p2.second->getEvents());
                int combined_usage = getUsageCounts().find(p1.first)->second + getUsageCounts().find(p2.first)->second;
                if (best_overlap.size() < overlap.size() || (best_overlap.size() == overlap.size() && best_combined_usage < combined_usage)){
                    best = p2.first;
                    best_overlap = overlap;
                    best_combined_usage = combined_usage;
                }
            }
        }
        if (best != -1){
            overlap_queue.emplace(p1.first, best, best_overlap, best_combined_usage);
        }
    }


    auto prev_encoded_length = encoded_length;
    while (!overlap_queue.empty()){
        auto next = overlap_queue.top();
        overlap_queue.pop();
        if (removed_pattern_ids.find(next.p2) != removed_pattern_ids.end() || removed_pattern_ids.find(next.p1) != removed_pattern_ids.end()) continue; // skip patterns merged with others

        /**
         *  1. create new needed abstractions
         *  2. create new pattern
         *  3. update cover (keep track of changes to roll back)
         *  4. recompute usages?
         *  5. recompute length
         *  6. if bits reduced leave as is and delete roll backs changes add p1 and p2 to already removed list
         *  else apply rollback to cover remove new pattern and abstractions
         */
         auto p1 = selection[next.p1];
         auto p2 = selection[next.p2];
         auto p1_events = p1->getEvents();
         auto p2_events = p2->getEvents();

         std::vector<int> added_abstractions; // for rollback

         std::vector<int> new_p_events;
         std::vector<int> dependent_on;
         std::vector<LinkType> link_type;
         for (int i =0; i < p1_events.size();i++){
             if (p1_events[i] == p2_events[i]){
                 new_p_events.push_back(p1_events[i]);
                 dependent_on.push_back(p1->getDependsOn()[i]);
                 link_type.push_back((p1->getLinkTypes()[i]));
             }else{
                 if (p1_events[i] > db.getHighestEvent() && p2_events[i] <= db.getHighestEvent() && inChildren(p1_events[i],p2_events[i])){
                    new_p_events.push_back(p1_events[i]);
                 }else if (p2_events[i] > db.getHighestEvent() && p1_events[i] <= db.getHighestEvent() && inChildren(p2_events[i], p1_events[i])){
                     new_p_events.push_back(p2_events[i]);
                 }
                 else{
                     Abstraction new_a = Abstraction(getNewAbstractionId(),{p1_events[i], p2_events[i]});
                     new_p_events.push_back(new_a.getId());
                     addAbstraction(new_a,false);
                     added_abstractions.push_back(new_a.getId());
                 }
                 dependent_on.push_back(-1);
                 link_type.push_back(DEPENDENT_ON);
             }
         }

         auto new_p  = new Pattern(new_p_events, dependent_on, link_type);
         addPatternToSelection(new_p, true);
         removePatternFromSelection(p1);
         removePatternFromSelection(p2);

         auto p1_locations  = p1->getLocations();
         auto p2_locations  = p2->getLocations();

         //update locations
         for (auto w:p1_locations){
             w->changePattern(new_p);
         }
         for (auto w:p2_locations){
             w->changePattern(new_p);
         }
         new_p->copyLocations();
         computePattenStructure(new_p);
         recomputeUsages();
         computeLength();
         if (prev_encoded_length < encoded_length){
             //ROLLBACK
             removePatternFromSelection(new_p);
             addPatternToSelection(p1, false);
             addPatternToSelection(p2, false);

             for (auto w:p1_locations){
                 w->changePattern(p1);
             }
             for (auto w:p2_locations){
                 w->changePattern(p2);
             }
             for (auto a:added_abstractions){
                 removeAbstraction(a);
             }
             recomputeUsages();
             computeLength();
         }else{
             removed_pattern_ids.insert(p1->getId());
             removed_pattern_ids.insert(p2->getId());
             availableWindowSet.changePatterns(p1,new_p);
             availableWindowSet.changePatterns(p2,new_p);
         }
         prev_encoded_length = encoded_length;
    }

    std::vector<int> abs_to_remove;
    for (auto a1:abstractions){
        auto children_a1 = getAllChildrenOfEvent(a1.first);
        std::sort(children_a1.begin(),children_a1.end());
        for (auto a2:abstractions){
            if (a1.first >= a2.first) continue;
            auto children_a2 = getAllChildrenOfEvent(a2.first);
            std::sort(children_a2.begin(), children_a2.end());
            if (children_a2.size() == children_a1.size()){
                bool match = true;
                for (int i =0; i < children_a1.size(); i++){
                    if (children_a1[i] != children_a2[i]){
                        match = false;
                        break;
                    }
                }
                if (match){
                    replaceSymbolInPatterns(a2.first,a1.first);
                    replaceSymbolInAbstractions(a2.first,a1.first);
                    abs_to_remove.push_back(a2.first);
                }
            }
        }
    }
    for (auto a:abs_to_remove){
        removeAbstraction(a);
    }

    std::vector<int> abs_ids;
    for (auto a_it = abstractions.begin(); a_it != abstractions.end();a_it++){
        abs_ids.push_back(a_it->first);
    }


    for (auto a_id:abs_ids){
        auto a_it = abstractions.find(a_id);
        std::vector<int> used_in;
        for (auto a2:abstractions){
            for (auto abs:a2.second.getChildren()){
                if ((*a_it).first == abs){
                    used_in.push_back(a2.first);
                }
            }
        }
        if (used_in.size() == 1){
            auto new_children = getAllChildrenOfEvent(a_it->first);
            auto more_children = getAllChildrenOfEvent(used_in[0]);
            new_children.insert(new_children.begin(),more_children.begin(),more_children.end());
            Abstraction new_a = Abstraction(getNewAbstractionId(),new_children);
            new_a.killTheDoppelgaengers();
            addAbstraction(new_a,false);

            auto pattern_changes1 = replaceSymbolInPatterns((*a_it).first,new_a.getId());
            auto pattern_changes2 = replaceSymbolInPatterns(used_in[0],new_a.getId());
            replaceSymbolInAbstractions(used_in[0], new_a.getId());
            removeElementFromAbstraction(used_in[0],(*a_it).first);
            Abstraction removed_abs1 = a_it->second;
            Abstraction removed_abs2 = abstractions.find(used_in[0])->second;
            auto a_it_if_removed  = abstractions.erase(a_it);
            abstractions.erase(abstractions.find(used_in[0]));
            double prev_length = encoded_length;
            recomputeUsages();
            computeLength();
            if (prev_length < encoded_length){
                extendAbstraction(used_in[0],{(*a_it).first});
                addAbstraction(removed_abs1, false);
                addAbstraction(removed_abs2, false);
                replaceSymbolInAbstractions(new_a.getId(), used_in[0]);
                abstractions.erase(abstractions.find(new_a.getId()));
                replaceBackPatterns(pattern_changes1, (*a_it).first);
                replaceBackPatterns(pattern_changes2, used_in[0]);
                recomputeUsages();
                computeLength();
            }else{
                abstractionRemoveSet.emplace_back((*a_it).first, new_a.getId());
                abstractionRemoveSet.emplace_back(used_in[0], new_a.getId());
            }

        }
    }

    recomputeUsages();
    computeLength();

    std::vector<int> removed_p_ids(removed_pattern_ids.begin(), removed_pattern_ids.end());
    return std::pair<std::vector<int>,std::vector<std::pair<int,int>>>(removed_p_ids,abstractionRemoveSet);
}

std::vector<int> Model::prunePatterns() {
    std::vector<int> removed_pattern_ids;
    std::vector<int> selection_keys;
    for (auto p:selection){
        selection_keys.push_back(p.first);
    }
    for (auto p_id:selection_keys){
        if (p_id <= db.getHighestEvent()) continue;
        auto prev_length = encoded_length;
        auto old_cover = cover;

        auto pattern = selection[p_id];
        for (auto it = cover.begin(); it != cover.end();){
            if (it->getPattern()->getId() == p_id){
                std::vector<Window> replacementCover;
                for (auto p:it->getPositions()){
                    replacementCover.push_back(Window(selection[db.getEvent(p)],{p}));
                }
                cover.insert(it,replacementCover.begin(),replacementCover.end());
                it = cover.erase(it);
            }else{
                it++;
            }
        }
        removePatternFromSelection(pattern);
        recomputeUsages();
        computeLength();
        if (prev_length < encoded_length){
            cover = old_cover;
            addPatternToSelection(pattern, false);
            recomputeUsages();
            computeLength();
        }else{
            removed_pattern_ids.push_back(p_id);
            availableWindowSet.removeWindowsOfPattern(p_id);
        }
    }
    return removed_pattern_ids;
}

std::vector<int> Model::pruneWithReCover() {

    std::vector<int> removed_pattern_ids;
    std::vector<int> selection_keys;
    for (auto p:selection){
        selection_keys.push_back(p.first);
    }
    for (auto p_id:selection_keys){
        if (p_id <= db.getHighestEvent()) continue;
        auto prev_length = encoded_length;
        auto old_cover = cover;
        auto pattern = selection[p_id];



        WindowSet windowSet = findWindows(*pattern,GlobalData::window_budget);
        removePatternFromSelection(pattern);
        availableWindowSet.removeWindowsOfPattern(p_id);
        cover = greedyCover();
        recomputeUsages();
        computeLength();
        if (prev_length < encoded_length){
            cover = old_cover;
            addPatternToSelection(pattern, false);
            availableWindowSet.addToWindowSet(windowSet);
            recomputeUsages();
            computeLength();
        }else{
            removed_pattern_ids.push_back(p_id);
        }
    }
    return removed_pattern_ids;
}

double computeCondEntropy(UniqueValueCounter<std::pair<int,int>> cont_table){
    int sum = 0;
    UniqueValueCounter<int> j_margin;
    for (auto c:cont_table.getCounts()){
        sum += c.second;
        j_margin.incCount(c.first.second,c.second);
    }
    double ent = 0;
    for (auto c:cont_table.getCounts()){
        double p_joined = (double)c.second/(double)sum;
        ent -= p_joined * log2(p_joined/((double)j_margin.getCount(c.first.second)/(double)sum));
    }
    return ent;
}

double computeSpecBitsNeeded(UniqueValueCounter<std::pair<int,int>> cont_table, double n_children){
    double episilon = 0.5;
    double bits =0;
    int i =1;
    for (auto c:cont_table.getCounts()){
        for (int j = 0; j < c.second ; j++){
            bits -= log2(episilon+j/(n_children * episilon + i + j));
            i++;
        }
    }
    return bits;
}

void Model::computePattenStructure(Pattern* pattern) {
    auto cont_table = pattern->getContingencyTableFromLocationCopy();
    for (int i =0; i < pattern->getEvents().size();i++){
        if (pattern->getEvents()[i] > db.getHighestEvent()){
            int n_children = getAllChildrenOfEvent(pattern->getEvents()[i]).size();
            double lowest_bits = MAXFLOAT;
            int best_j = -1;
            for (int j =-1; j < i; j++){
                auto k_it = cont_table.find(std::pair<int,int>(i,j));
                if (k_it != cont_table.end()){
                    double bits = computeSpecBitsNeeded(k_it->second, n_children);
                    if (bits < lowest_bits){
                        lowest_bits = bits;
                        best_j = j;
                    }
                }
            }
            pattern->setDependentOn(i,best_j);
        }
    }
}

std::vector<std::pair<int,int>> Model::replaceSymbolInPatterns(int to_replace,int with) {
    return replaceSymbol(selection, to_replace,with);
}

void Model::replaceBackPatterns(std::vector<std::pair<int,int>> changes, int with){
    for (auto p: changes){
        selection.find(p.first)->second->replaceEvent(with, p.second);
    }
}
void Model::replaceSymbolInAbstractions(int to_replace,int with){
    for (auto& a:abstractions){
        a.second.replaceSymbol(to_replace,with);
        a.second.killTheDoppelgaengers();
    }
}

bool Model::inChildren(int abs, int event) {
    auto all_children = getAllChildrenOfEvent(abs);
    for (auto child:all_children){
        if (event == child){
            return true;
        }
    }
    return false;
}

void Model::replaceSymbolInOpen(int to_replace, int with) {
    replaceSymbol(open,to_replace,with);
}

std::vector<std::pair<int,int>> Model::replaceSymbol(patternSet& ps, int to_replace, int with) {
    std::vector<std::pair<int,int>> replaced;
    for (auto p:ps){
        auto p_events = p.second->getEvents();
        for (int i =0; i< p_events.size();i++){
            if (p_events[i] == to_replace){
                replaced.push_back(std::pair<int,int>(p.first,i));
                p.second->replaceEvent(with,i);
            }
        }
    }
    return replaced;
}


void Model::removeUnusedAbstractions() {
    for (auto it = abstractions.begin(); it != abstractions.end();){
        if (abstractionUsed(it->first)){
            it++;
        }else{
            it = abstractions.erase(it);
        }
    }
}

bool Model::abstractionUsed(int id) {
    for (auto p:selection){
        for (auto e:p.second->getEvents()){
            for (int a:getAllAbstractionsOfEvent(e)){
                if (a == id){
                    return true;
                }
            }
        }
    }
    return false;
}

void Model::reCoverData() {
    cover = greedyCover();
    recomputeUsages();
    computeLength();
}

int Model::checkCover() {
    std::map<int,int> singleton_counts;
    for (auto w:cover){
        for (auto p:w.getPositions()){
            int event = db.getEvent(p);
            auto e_it = singleton_counts.find(event);
            if (e_it == singleton_counts.end()){
                singleton_counts[event] = 1;
            }else{
                e_it->second++;
            }
        }
    }
    int r_value  = -1;
    for (auto e1:usage_copy_diagnositics){
        if (singleton_counts.find(e1.first)->second != e1.second){
            std::cout << "DIFFERENCE for " << e1.first << " = " << singleton_counts.find(e1.first)->second - e1.second << std::endl;
            r_value =  e1.first;
        }
    }
    return r_value;
}

void Model::extendAbstractions() {
    /*
     * Create wildcard (simple abstraction incl. all elements)
     * Replace abstractions with wildcard
     * use wildcard matches
     */
    std::vector<int> all_children = std::vector<int>(db.getHighestEvent()+1);
    for (int i = 0; i < all_children.size(); i++){
        all_children[i] = i;
    }
    AbstractionExtension wildcard = AbstractionExtension(ADD_NEW, all_children, ++abstractionCounter);
    for (auto p:selection){
        if (p.first <= db.getHighestEvent()) continue;
        Pattern copy = *p.second;
        auto events = copy.getEvents();
        for (int i =0; i < events.size();i++){
            if (events[i]> db.getHighestEvent()){
                auto all_children_in_abs = getAllChildrenOfEvent(events[i]);
                copy.replaceEvent(wildcard.id,i);
                auto windows = findWindows(copy,GlobalData::window_budget,wildcard,{});
                UniqueValueCounter<int> wildcard_usage_counter;
                for (auto w:windows){
                    wildcard_usage_counter.incCount(db.getEvent(w.getPositions()[i]),1);
                }
                wildcard_usage_counter.print();
                auto counts = wildcard_usage_counter.getSortedCounts();
                std::vector<AbstractionExtension> abs_ext_list;
                double max_gain = 0;
                for (auto c:counts){
                    bool already_in = false;
                    for (auto abs_child:all_children_in_abs){
                        if (c.key == abs_child){
                            already_in = true;
                            break;

                        }
                    }
                    if (!already_in){
                        auto abs_ext = AbstractionExtension(EXTEND,{c.key}, events[i]);
                        abs_ext_list.push_back(abs_ext);
                        auto gain = computeGainAbsExt( abs_ext_list, *p.second);
                        if (gain > max_gain){
                            max_gain = gain;
                        }else{
                            abs_ext_list.erase(--abs_ext_list.end());
                            break;
                        }
                    }
                }
                applyAbstractionExtensionList(abs_ext_list);
                recomputeUsages();
                computeLength();
            }
        }
    }
}

ExtendCandidateReturn Model::cutOutAbstractions(ExtendCandidateReturn candidate) {

    std::vector<int> events_pattern;
    std::vector<int> dep_on;
    std::vector<LinkType> linkType;
    for (auto e:candidate.best_extension->getEvents()){
        if (e <= GlobalData::db->getHighestEvent()){
            events_pattern.push_back(e);
            dep_on.push_back(-1);
            linkType.push_back(NONE);
        }
    }
    auto new_p = new Pattern(++patternCounter,events_pattern,dep_on,linkType);
    return ExtendCandidateReturn(0, candidate.source_pattern, new_p, {});
}








