//
// 
//

#include "Pattern.h"
#include "GlobalData.h"

#include <utility>
#include <vector>
#include <iostream>


Pattern::Pattern(const std::vector<int>& events, std::vector<int> dependent_on, std::vector<LinkType> link_type) {
    this->events = events;
    this->depends_on = std::move(dependent_on);
    this->link_type = std::move(link_type);
    usageCounts = std::vector<std::map<std::pair<int,int>,int>>(events.size());
}

Pattern::Pattern(int id, std::vector<int> events, std::vector<int> dependent_on, std::vector<LinkType> link_type) : Pattern(events,dependent_on,link_type) {
    this->id = id;
}


std::list<Window*>& Pattern::getLocations() {
    return locations;
}

std::list<Window>& Pattern::getCopyLocations() {
    return copy_location;
}

int Pattern::getPatternUsageCounts() {
    return locations.size();
}

void Pattern::addLocation(Window* w) {
    locations.push_back(w);
}



std::vector<int> Pattern::getEvents() const{
    return events;
}

int Pattern::getId() {
    return id;
}

void Pattern::setId(int id) {
    this->id = id;
}

int Pattern::dependentOn(int i) const{
    return depends_on[i];
}

LinkType Pattern::kindOfDependence(int i) const{
    return link_type[i];
}


void Pattern::setDependentOn(int i, int dependent_on) {
    depends_on[i] = dependent_on;
}

void Pattern::clearUsageCounts() {
    usageCounts = std::vector<std::map<std::pair<int,int>,int>>(events.size());
}

void Pattern::incUsageCounts(int i, int conditionedOn, int event) {
    if (usageCounts[i].find({conditionedOn,event}) != usageCounts[i].end()){
        usageCounts[i][{conditionedOn,event}]++;
    }else{
        usageCounts[i][{conditionedOn,event}] = 1;
    }
}

int Pattern::getUsageCount(int i, int conditionedOn, int event) {
    if (usageCounts[i].find({conditionedOn,event}) != usageCounts[i].end()){
        return usageCounts[i][{conditionedOn,event}];
    }else{
        return 0;
    }
}

int Pattern::getSumUsageCount(int i, int conditionedOn) {
    int sum =0;
    for (auto count:usageCounts[i]){
        if (count.first.first == conditionedOn){
            sum += count.second;
        }
    }
    return sum;
}


UniqueValueCounter<int> Pattern::getUnconditionalUsageCounts(int i) {
    UniqueValueCounter<int> counts;
    for (auto condionalCounts:usageCounts[i]){
        counts.incCount(condionalCounts.first.second,condionalCounts.second);
    }
    return counts;
}

void Pattern::clearLocations() {
    locations.clear();
}


void Pattern::addWindow(Window w) {
    allWindows.push_back(w);

    auto covered = w.getPositions();



    if (covered.size() != events.size()){ // sanity check
        throw std::runtime_error("Covered Events do not match pattern length");
    }

    for (int i = 0; i < covered.size();i++){
        margin_usage_counts[i].incCount(GlobalData::db->getEvent(covered[i]),1);
    }

    auto e_it = events.begin();
    auto c_e_it = covered.begin();
    while (e_it != events.end()){
        if (*e_it > GlobalData::db->getHighestEvent()){// if abstraction
            auto e_covered = GlobalData::db->getEvent(*c_e_it);

            auto next_abs = e_it;
            auto c_next_abs = c_e_it;
            next_abs++;
            c_next_abs++;
            while (next_abs != events.end() ){
                if (*next_abs > GlobalData::db->getHighestEvent()){
                    auto next_abs_covered = GlobalData::db->getEvent(*c_next_abs);
                    auto key = std::pair<int,int>(*e_it,*next_abs);
                    auto k_it = contingency_tables_over_all_windows.find(key);
                    if (k_it != contingency_tables_over_all_windows.end()){
                        k_it->second.incCount(std::pair<int,int>(e_covered,next_abs_covered),1);
                    }else{
                        contingency_tables_over_all_windows[key] = UniqueValueCounter<std::pair<int,int>>();
                        contingency_tables_over_all_windows[key].incCount(std::pair<int,int>(e_covered,next_abs_covered),1);
                    }

                }
                next_abs++;
                c_next_abs++;
            }
        }
        e_it++;
        c_e_it++;
    }
}

void Pattern::clearWindows_and_Contingency_tables(){
    allWindows.clear();
    contingency_tables_over_all_windows.clear();
    margin_usage_counts = std::vector<UniqueValueCounter<int>>(events.size());
    allWindowsUpToDate = false;
}

void Pattern::allWindowsSet(){
    allWindowsUpToDate = true;
}

void Pattern::voidWindowSet(){
    allWindowsUpToDate = false;
}

int Pattern::allWindowsSize(){
    return allWindows.size();
}

int Pattern::getMarginCounts(int i, int event){
    return margin_usage_counts[i].getCount(event);
}


bool Pattern::operator==(const Pattern &rhs) const {
    for (int i =0;i<events.size();i++){
        if (events[i] != rhs.events[i] ||
        depends_on[i] != rhs.depends_on[i] ||
        link_type[i] != link_type[i]){
            return false;
        }
    }
    return true;
}

bool Pattern::operator!=(const Pattern &rhs) const{
    return !(*this == rhs);
}
bool Pattern::operator>(const Pattern &rhs) const {
    return (this->getEvents().size() > rhs.getEvents().size());
}

void Pattern::print() {
    std::cout << "Id: " << id << std::endl;
    for (auto e:events){
        std::cout << e << '\t';
    }
    std::cout << std::endl;
    for (auto d:depends_on){
        std::cout << d << '\t';
    }
    std::cout << std::endl;
    for (auto l:link_type){
        std::cout << l << '\t';
    }
    std::cout << std::endl;
}

void Pattern::replace(int position, int event, int dependent_on, LinkType linkType, bool splice) { 
    allWindowsUpToDate = false;
     if (position > -1 && position <= events.size()){
        if (splice){ 
            auto ev_position_it = events.begin();
            std::advance(ev_position_it,position);
            events.insert(ev_position_it,event);
            auto dp_position_it = depends_on.begin();
            std::advance(dp_position_it,position);
            depends_on.insert(dp_position_it,dependent_on);
            auto ly_pos_it = link_type.begin();
            std::advance(ly_pos_it,position);
            link_type.insert(ly_pos_it,linkType);
        }else{ // else we replace
            events[position] = event;
            depends_on[position] = dependent_on;
            link_type[position] = linkType;
        }
    }else{
        throw std::out_of_range("Position not in pattern!");
    }
}

void Pattern::splice(int position, int event, int dependent_on, LinkType linkType) {

}

std::vector<int> Pattern::getDependsOn() {
    return depends_on;
}

std::vector<LinkType> Pattern::getLinkTypes() {
    return link_type;
}

void Pattern::extendWithPattern(Pattern &pattern, int position) {
    allWindowsUpToDate = false;
    if (position >= 0 && position <= events.size()){
        auto ev_position_it = events.begin();
        std::advance(ev_position_it,position);
        auto patternEvents = pattern.getEvents();
        events.insert(ev_position_it,patternEvents.begin(),patternEvents.end());

        auto dp_position_it = depends_on.begin();
        std::advance(dp_position_it,position);
        auto pattenDependsOn = pattern.getDependsOn();
        depends_on.insert(dp_position_it,pattenDependsOn.begin(), pattenDependsOn.end());

        auto lt_position_it = link_type.begin();
        std::advance(lt_position_it,position);
        auto patterLinkType = pattern.getLinkTypes();
        link_type.insert(lt_position_it,patterLinkType.begin(),patterLinkType.end());
    }else{
        throw std::out_of_range("Position not in pattern!");
    }
}

void Pattern::copyLocations() {
    copy_location.clear();
    for (auto w:locations){
        copy_location.push_back(*w);
    }
}

contingency_table Pattern::getContingencyTableFromLocationCopy() {

    contingency_table cont_table;
    for (int i = 0; i < this->events.size(); i++){
        if (this->events[i] > GlobalData::db->getHighestEvent()){
            for (int j =-1; j < i; j++){
                if (this->events[j] > GlobalData::db->getHighestEvent()){
                    cont_table[std::pair<int,int>(i,j)] = UniqueValueCounter<std::pair<int,int>>();
                }
            }
        }
    }


    for (auto w:copy_location){
        auto cov_positions = w.getPositions();
        for (auto &c:cont_table){
            auto i_cov = GlobalData::db->getEvent(cov_positions[c.first.first]);
            int j_cov;
            if (c.first.second == -1){
                j_cov = -1;
            }else{
                j_cov = GlobalData::db->getEvent(cov_positions[c.first.second]);
            }
            c.second.incCount(std::pair<int,int>(i_cov,j_cov),1);
        }
    }
    return cont_table;
}

void Pattern::replaceEvent(int new_event, int index) {
    events[index] = new_event;
}

void Pattern::removeEvent(int index){
    clearLocations();
    clearUsageCounts();
    clearWindows_and_Contingency_tables();
    auto e_it = events.begin();
    std::advance(e_it,index);
    events.erase(e_it);

    auto d_it = depends_on.begin();
    auto l_it = link_type.begin();
    std::advance(d_it, index);
    std::advance(l_it, index);
    d_it = depends_on.erase(d_it);
    l_it = link_type.erase(l_it);
    while (d_it != depends_on.end()){
        if (*d_it > index) (*d_it)--;
        if (*d_it == index){
            *d_it =-1;
            *l_it = NONE;
        }
        d_it++;
        l_it++;
    }

}

std::string Pattern::getOutputString() {
    std::string output = "";
    for (auto e:events){
        output += std::to_string(e) + " ";
    }
    output +=  "-2 ";
    for (auto d:depends_on){
        if (d == -1) d = -3;
        output += std::to_string(d) + " ";
    }
    output += "-2 ";
    for (auto l:link_type){
        int ld;
        switch (l) {
            case NONE:
                ld = 0;
                break;
            case EQUALS:
                ld = 1;
                break;
            case DEPENDENT_ON:
                ld = 2;
                break;
            default:
                throw std::runtime_error("Should not be the case!");
        }
        output += std::to_string(ld) + " ";
    }
    output += "-1";
    return output;
}

std::string Pattern::getTransPasString(){
    std::string output = "";
    auto cont_table = getContingencyTableFromLocationCopy();
    for (int i =0; i < events.size();i++ ){
        output += "{ ";
        if (events[i] > GlobalData::db->getHighestEvent()){
            if (depends_on[i] != -1){
                auto k_it = cont_table.find(std::pair<int,int>(i,depends_on[i]));
                for (auto count: k_it->second.getCounts()){
                    auto f = std::to_string(count.first.first);
                    auto t = std::to_string(count.first.second);
                    auto c = std::to_string(count.second);
                    output += " (" + f + "," + t + "):" + c;
                }
            }
            else{
                for (auto c :this->getUnconditionalUsageCounts(i).getCounts()){
                    output += " (" + std::to_string(c.first) + "):" + std::to_string(c.second);
                }
            }
        }
        output += "}, ";
    }
    return output;
}


std::vector<int> Abstraction::getChildren() {
    this->killTheDoppelgaengers();
    return children;
}

Abstraction::Abstraction(int id, std::vector<int> children): Abstraction(children){
    this->id = id;
}

Abstraction::Abstraction(std::vector<int> children): children(children) {}

int Abstraction::getId() {
    return id;
}

void Abstraction::setId(int id) {
    this->id = id;
}

void Abstraction::extendAbstraction(int extension) {
    children.push_back(extension);
}

void Abstraction::removeChildrenFromAbstraction(std::vector<int> removeSet) {
    //SHOULD DO more efficient implementation
    for (auto it = children.begin(); it != children.end();){
        bool erased = false;
        for (int e:removeSet){
            if (*it == e){
                it = children.erase(it);
                erased = true;
                break;
            }
        }
        if (!erased){
            ++it;
        }
    }
}

void Abstraction::killTheDoppelgaengers() {
    std::set<int> seen;
    for (auto it = children.begin(); it != children.end();){
        if (seen.find(*it) != seen.end()){
            it = children.erase(it);
        }else{
            seen.insert(*it);
            it++;
        }
    }
}

void Abstraction::replaceSymbol(int to_replace, int with) {
    for (int i = 0; i < children.size(); i++){
        if (children[i] == to_replace){
            children[i] = with;
        }
    }
}

std::string Abstraction::getOutputString() {
    std::string output;
    auto local_children = this->getChildren();
    std::sort(local_children.begin(), local_children.end());
    for (auto c:local_children){
        output += std::to_string(c) + " ";
    }
    output += "-1 ";
    return output;
}

bool Abstraction::operator<(const Abstraction &rhs) const {
    return (this->id < rhs.id);
}

Candidate::Candidate(Pattern &pattern, int position, extension_list extensionList): pattern(pattern), position(position), extensionList(extensionList) {}

extension_list Candidate::getExtensions() {
    return extensionList;
}



int Candidate::getPosition() {
    return position;
}

Pattern* Candidate::getCopyOfPattern() {
    return new Pattern(pattern.getEvents(), pattern.getDependsOn(), pattern.getLinkTypes());
}

Pattern& Candidate::getPattern() {
    return pattern;
}

void Candidate::addExtensionToList(int id, GapDistribution gp) {
    extensionList[id] = gp;
}

void Candidate::printExtensionList() {
    std::cout << "Extension List: " << std::endl;
    for (auto e:extensionList){
        std::cout << e.first << " - " << e.second.getTotalCounts() << std::endl;
    }
    std::cout << std::endl;
}
