//
// 
//

#include <algorithm>

#include "utils.h"
#include "iostream"


double universal_integer_encoding(int i) {
    if (i < 1){
        throw  std::invalid_argument("Universal Integer Encoding only defined for values larger 0");
    }
    double c = 2.865064;
    double bits = log2(c);
    double d_i =  (double)i;
    d_i = log2(d_i);
    while (d_i >0){
        bits += d_i;
        d_i = log2(d_i);
    }
    return bits;
}

// locFact and logBinom implemented by Jonas Fischer
static std::vector<double> logFacStorage (1,0);
double logFact(const uint32_t n){
    // evaluate everything up to requested value
    for (uint32_t i = logFacStorage.size(); i <= n; ++i)
    {
        logFacStorage.push_back(logFacStorage[i - 1] + log2(static_cast<double>(i)));
    }
    return logFacStorage[n];
}
// return log2 of binomial coefficient n choose k
double logBinom(const uint32_t n, const uint32_t k){
    if (k > n){
        std::cerr << "logBinom - n smaller k" << std::endl;
        return 0;
//        throw std::runtime_error("logBinom - n smaller k");
    }

    return logFact(n) - logFact(n-k) - logFact(k);
}

bool checkOverlap(std::vector<DataIndex>& p1, std::vector<DataIndex>& p2){
    auto p1It = p1.begin();
    auto p2It = p2.begin();
    while (p1It != p1.end() && p2It != p2.end()){
        if (*p1It == *p2It){
            return true;
        }
        if (*p1It < *p2It){
            p1It++;
        } else{
            p2It++;
        }
    }
    return false;
}

void GapDistribution::incDeltaFrequency(int delta, int value) {
    m_stat_correct = false; //m_stat has to recomputed

    auto distributionIt = distribution.find(delta);
    if (distributionIt == distribution.end()){
        distribution[delta] = value;
    }else{
        distribution[delta] += value;
    }
    total_counts += value;
}

int GapDistribution::getDeltaFrequency(int delta) {
    auto distributionIt = distribution.find(delta);
    if (distributionIt == distribution.end()){
        throw std::runtime_error("Key not found");
    }else{
        return distribution[delta];
    }
}

double GapDistribution::getDeltaProbability(int delta) {
    return (double)getDeltaFrequency(delta)/(double)total_counts;
}

void GapDistribution::removeDeltaFromDistribution(int delta) {
    m_stat_correct = false;

    auto distributionIt = distribution.find(delta);
    if (distributionIt != distribution.end()){
        total_counts -= distribution[delta];
        distribution.erase(distributionIt);
    }else{
        throw std::runtime_error("Key not found");
    }
}

std::vector<int> GapDistribution::getAllDeltas() {
    std::vector<int> keys;
    for (auto it = distribution.begin(); it != distribution.end(); ++it){
        keys.push_back(it->first);
    }
    return keys;
}

int GapDistribution::getTotalCounts() {
    return total_counts;
}

double GapDistribution::divergence(GapDistribution &other) {
    // call devergence measure
    return earth_move_distance(other);
}

double GapDistribution::divergence_frqu_adjsted(GapDistribution &other) {
    return divergence(other)/other.getTotalCounts();
}

double GapDistribution::kullback_leibler_discrete(GapDistribution &other) {
    auto deltas = this->getAllDeltas();
    std::sort(deltas.begin(), deltas.end());
    double sum = 0;
    for (auto delta:deltas){
        auto p_this = this->getDeltaProbability(delta);
        auto p_other = other.getDeltaProbability(delta); // will throw an error if not in other
        sum += p_this * log2(p_this/p_other);
    }
    return sum;
}

double kullback_leibler_discrete(std::vector<double> p, std::vector<double> m){
    auto p_it = p.begin();
    auto m_it = m.begin();
    double  sum = 0;
    while (m_it != m.end()){
        if (*p_it != 0.0){
            sum += *p_it * log2(*p_it/ *m_it);
        }
        p_it++;
        m_it++;
    }
    return sum;
}

double GapDistribution::jensen_shannon_discrete(GapDistribution &other) {
    std::vector<double> m;
    std::vector<double> p;
    std::vector<double> q;
    auto this_deltas = this->getAllDeltas();
    std::sort(this_deltas.begin(),this_deltas.end());
    auto other_deltas = other.getAllDeltas();
    std::sort(other_deltas.begin(),other_deltas.end());
    auto this_deltas_it = this_deltas.begin();
    auto other_deltas_it = other_deltas.begin();
    while (this_deltas_it != this_deltas.end() && other_deltas_it != other_deltas.end()){
        if (*this_deltas_it == *other_deltas_it){
            m.push_back(.5 * (this->getDeltaProbability(*this_deltas_it) + other.getDeltaProbability(*other_deltas_it)));
            p.push_back(this->getDeltaProbability(*this_deltas_it));
            q.push_back(other.getDeltaProbability(*other_deltas_it));
            this_deltas_it++;
            other_deltas_it++;
        }else if (*this_deltas_it < *other_deltas_it){
            m.push_back(.5 * this->getDeltaProbability(*this_deltas_it));
            p.push_back(this->getDeltaProbability(*this_deltas_it));
            q.push_back(0.0);
            this_deltas_it++;
        }else if (*other_deltas_it < *this_deltas_it){
            m.push_back(.5 * other.getDeltaProbability(*other_deltas_it));
            p.push_back(0.0);
            q.push_back(other.getDeltaProbability(*other_deltas_it));
            other_deltas_it++;
        } else{
            throw std::runtime_error("How did this happen?");
        }
    }
    return .5 * ::kullback_leibler_discrete(p, m) + .5 * ::kullback_leibler_discrete(q, m);
}



void GapDistribution::print() {
    for (auto e:distribution){
        std::cout << e.first << " - " << e.second << std::endl;
    }
}
/*
 * returns either mean, mode or median
 */
double GapDistribution::getMStat() {
    if (!m_stat_correct){
        m_stat = getMedian(); //replace with desired method
        m_stat_correct = true;
    }
    return m_stat;
}

double GapDistribution::getMean() {
    double sum = 0;
    for (auto id_count:distribution){
        sum += id_count.first * id_count.second;
    }
    return sum/total_counts;
}

double GapDistribution::getMode() {
    int mode;
    int max_frequ = 0;
    for (auto e:distribution){
        if (e.second > max_frequ){
            mode = e.first;
            max_frequ = e.second;
        }
    }
    return (double) mode;
}

double GapDistribution::getMedian() {
    int sum =0;
    for (auto e:distribution){
        sum += e.second;
        if (sum >= total_counts/2){
            return e.first;
        }
    }
    if (total_counts == 0){
        return 0;
    }
    throw std::runtime_error("Can't be");
}

void splitIntoVectors(std::map<int,int>& map, std::vector<double>& key_v, std::vector<double>& value_v){
    for (auto e:map){
        key_v.push_back(e.first);
        value_v.push_back(e.second);
    }
}

double GapDistribution::earth_move_distance(GapDistribution &other) {
    std::vector<double> this_values;
    std::vector<double> this_weights;
    std::vector<double> other_values;
    std::vector<double> other_weights;
    splitIntoVectors(this->distribution,this_values,this_weights);
    splitIntoVectors(other.distribution, other_values, other_weights);
    return std::wasserstein(this_values, this_weights, other_values, other_weights);
}


